###########################################################################
#
#  Copyright 2020 Google LLC
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      https://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
#
###########################################################################

from datetime import date, datetime, timedelta
import re
import random

from googleapiclient.errors import HttpError

from starthinker.util.bigquery import json_to_table
from starthinker.util.bigquery import query_to_view
from starthinker.util.bigquery import rows_to_table
from starthinker.util.bigquery import table_to_rows
from starthinker.util.google_api import API_BigQuery
from starthinker.util.project import project

RE_EMAIL = re.compile(r'(^[a-zA-Z0-9_.+-]+@[a-zA-Z0-9-]+\.[a-zA-Z0-9-.]+$)')
INTEGER_MULTIPLY = random.randint(2, 9)
INTEGER_OFFSET = random.randint(17, 99)
DATE_OFFSET = timedelta(days=random.randint(3, 13), weeks=random.randint(3, 23))

STRING_IDS = {}


def anonymize_string(cell, column):
  global STRING_IDS
  STRING_IDS.setdefault(column, {})
  STRING_IDS[column].setdefault(cell, len(STRING_IDS[column]) + 1)

  # email
  if RE_EMAIL.match(cell):
    return '%s_%d@email.com' % (column.rsplit('.',
                                              1)[-1], STRING_IDS[column][cell])
  # any string
  else:
    return '%s_%d' % (column.rsplit('.', 1)[-1], STRING_IDS[column][cell])


def anonymize_integer(cell, column):
  # possible date ( legacy Data Studio format )
  if len(str(cell)) == 4 + 2 + 2:
    return cell
  else:
    # integer
    return (cell * INTEGER_MULTIPLY) + INTEGER_OFFSET


def anonymize_float(cell, column):
  # replace integers randonly within float to preserve semaintics ( in case 0 - 1 range is percent )
  return float(''.join([
      '.' if char == '.' else str(random.randint(0, 9))
      for char in str(cell).lstrip('0')
  ]))


def anonymize_date(cell, column):
  # shift date back in time to prevent future dates
  # convert to string because its being written back to BigQuery
  return str(cell - DATE_OFFSET)


def anonymize_value(value, column):
  if isinstance(value, int):
    return anonymize_integer(value, column)
  elif isinstance(value, str):
    return anonymize_string(value, column)
  elif isinstance(value, float):
    return anonymize_float(value, column)
  elif isinstance(value, date):
    return anonymize_date(value, column)
  elif isinstance(value, datetime):
    return anonymize_date(value, column)


def anonymize_json(struct, parent=None):
  if isinstance(struct, dict):
    for key, value in struct.items():
      if isinstance(value, (dict, list)):
        anonymize_json(value, '.'.join((parent, key)) if parent else key)
      else:
        struct[key] = anonymize_value(
            value, '.'.join((parent, key)) if parent else key)
  elif isinstance(struct, list):
    for index, value in enumerate(struct):
      if isinstance(value, (dict, list)):
        anonymize_json(value, parent)
      else:
        struct[index] = anonymize_value(value, parent)

  return struct


def anonymize_csv(row, schema):
  for index, cell in enumerate(row):
    if cell is not None:
      if schema[index]['type'] == 'STRING':
        row[index] = anonymize_string(cell, schema[index]['name'])
      elif schema[index]['type'] in ('INTEGER', 'INT64', 'NUMERIC'):
        row[index] = anonymize_integer(cell, schema[index]['name'])
      elif schema[index]['type'] in ('FLOAT', 'FLOAT64'):
        row[index] = anonymize_float(cell, schema[index]['name'])
      elif schema[index]['type'] in ('DATE', 'DATETIME'):
        row[index] = anonymize_date(cell, schema[index]['name'])
  return row


def anonymize_rows(rows, schema, is_object):
  for row in rows:
    if is_object:
      yield anonymize_json(row)
    else:
      yield anonymize_csv(row, schema)


def anonymize_table(table_id):

  if project.verbose:
    print('ANONYMIZE TABLE', project.task['bigquery']['to']['dataset'],
          table_id)

  schema = API_BigQuery(project.task['auth']).tables().get(
      projectId=project.task['bigquery']['from']['project'],
      datasetId=project.task['bigquery']['from']['dataset'],
      tableId=table_id).execute()['schema']['fields']

  is_object = any([s['type'] == 'RECORD' for s in schema])

  rows = table_to_rows(
      project.task['auth'],
      project.task['bigquery']['from']['project'],
      project.task['bigquery']['from']['dataset'],
      table_id,
      as_object=is_object)

  rows = anonymize_rows(rows, schema, is_object)

  if is_object:
    json_to_table(
        project.task['auth'],
        project.task['bigquery']['to']['project'],
        project.task['bigquery']['to']['dataset'],
        table_id,
        rows,
        schema,
        disposition='WRITE_TRUNCATE')
  else:
    rows_to_table(
        project.task['auth'],
        project.task['bigquery']['to']['project'],
        project.task['bigquery']['to']['dataset'],
        table_id,
        rows,
        schema,
        skip_rows=0,
        disposition='WRITE_TRUNCATE')


def copy_view(view_id):
  if project.verbose:
    print('ANONYMIZE VIEW', project.task['bigquery']['to']['dataset'], view_id)

  view = API_BigQuery(project.task['auth']).tables().get(
      projectId=project.task['bigquery']['from']['project'],
      datasetId=project.task['bigquery']['from']['dataset'],
      tableId=view_id).execute()['view']

  project_dataset_template = '[%s:%s.' if view['useLegacySql'] else '`%s.%s.'

  query = view['query'].replace(
      project_dataset_template % (project.task['bigquery']['from']['project'],
                                  project.task['bigquery']['from']['dataset']),
      project_dataset_template % (project.task['bigquery']['to']['project'],
                                  project.task['bigquery']['to']['dataset']))

  query_to_view(
      project.task['auth'],
      project.task['bigquery']['to']['project'],
      project.task['bigquery']['to']['dataset'],
      view_id,
      query,
      legacy=view['useLegacySql'],
      replace=True)


@project.from_parameters
def anonymize():

  views = []

  for table in API_BigQuery(
      project.task['auth'], iterate=True).tables().list(
          projectId=project.task['bigquery']['from']['project'],
          datasetId=project.task['bigquery']['from']['dataset']).execute():
    if table['type'] == 'VIEW':
      views.append(table['tableReference']['tableId'])
    else:
      anonymize_table(table['tableReference']['tableId'])

  # views have dependencies, loop through all and create until no more errors or no change in view list
  last_copy = True
  while last_copy:
    retry_views = []
    last_copy = False
    while views:
      view = views.pop()
      try:
        copy_view(view)
        last_copy = True
      except HttpError as e:
        if e.resp.status == 404:
          retry_views.append(view)
        else:
          raise e
    views = retry_views


if __name__ == '__main__':
  anonymize()



###########################################################################
#
#  Copyright 2020 Google LLC
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      https://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
#
###########################################################################

'''
--------------------------------------------------------------

Before running this Airflow module...

  Install StarThinker in cloud composer ( recommended ):

    From Release: pip install starthinker
    From Open Source: pip install git+https://github.com/google/starthinker

  Or push local code to the cloud composer plugins directory ( if pushing local code changes ):

    source install/deploy.sh
    4) Composer Menu
    l) Install All

--------------------------------------------------------------

  If any recipe task has "auth" set to "user" add user credentials:

    1. Ensure an RECIPE['setup']['auth']['user'] = [User Credentials JSON]

  OR

    1. Visit Airflow UI > Admin > Connections.
    2. Add an Entry called "starthinker_user", fill in the following fields. Last step paste JSON from authentication.
      - Conn Type: Google Cloud Platform
      - Project: Get from https://github.com/google/starthinker/blob/master/tutorials/cloud_project.md
      - Keyfile JSON: Get from: https://github.com/google/starthinker/blob/master/tutorials/deploy_commandline.md#optional-setup-user-credentials

--------------------------------------------------------------

  If any recipe task has "auth" set to "service" add service credentials:

    1. Ensure an RECIPE['setup']['auth']['service'] = [Service Credentials JSON]

  OR

    1. Visit Airflow UI > Admin > Connections.
    2. Add an Entry called "starthinker_service", fill in the following fields. Last step paste JSON from authentication.
      - Conn Type: Google Cloud Platform
      - Project: Get from https://github.com/google/starthinker/blob/master/tutorials/cloud_project.md
      - Keyfile JSON: Get from: https://github.com/google/starthinker/blob/master/tutorials/cloud_service.md

--------------------------------------------------------------

CM User Editor

A tool for rapidly bulk editing Campaign Manager profiles, roles, and sub accounts.

  - Add this card to a recipe and save it.
  - Then click <strong>Run Now</strong> to deploy.
  - Follow the <a href="https://docs.google.com/spreadsheets/d/1zQDqm3owQMNsx0lvcN1SQWVGOhwfgsWSmirBK54qVL0/">instructions</a> for setup.

--------------------------------------------------------------

This StarThinker DAG can be extended with any additional tasks from the following sources:
  - https://google.github.io/starthinker/
  - https://github.com/google/starthinker/tree/master/dags

'''

from starthinker.airflow.factory import DAG_Factory

INPUTS = {
  'recipe_name': '',  # Name of document to deploy to.
}

RECIPE = {
  'setup': {
    'day': [
    ],
    'hour': [
    ]
  },
  'tasks': [
    {
      'drive': {
        'auth': 'user',
        'hour': [
        ],
        'copy': {
          'source': 'https://docs.google.com/spreadsheets/d/1zQDqm3owQMNsx0lvcN1SQWVGOhwfgsWSmirBK54qVL0/',
          'destination': {
            'field': {
              'name': 'recipe_name',
              'prefix': 'CM User Editor For ',
              'kind': 'string',
              'order': 1,
              'description': 'Name of document to deploy to.',
              'default': ''
            }
          }
        }
      }
    }
  ]
}

DAG_FACTORY = DAG_Factory('cm_user_editor', RECIPE, INPUTS)
DAG = DAG_FACTORY.generate()

if __name__ == "__main__":
  DAG_FACTORY.print_commandline()



###########################################################################
#
#  Copyright 2020 Google LLC
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      https://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
#
###########################################################################
"""Handles creation and updates of creative assets."""

import os
import json
import mimetypes
from io import BytesIO

from starthinker.config import BUFFER_SCALE
from starthinker.util.storage import object_get, object_get_chunks, MediaIoBaseUpload
from starthinker.task.traffic.dao import BaseDAO
from starthinker.task.traffic.feed import FieldMap
from starthinker.task.traffic.store import store

CHUNKSIZE = int(200 * 1024000 *
                BUFFER_SCALE)  # scale is controlled in config.py


class CreativeAssetDAO(BaseDAO):
  """Creative asset data access object.

  Inherits from BaseDAO and implements ad specific logic for creating and
  updating ads.
  """

  def __init__(self, auth, profile_id, is_admin, gc_project):
    """Initializes CreativeAssetDAO with profile id and authentication scheme."""
    super(CreativeAssetDAO, self).__init__(auth, profile_id, is_admin)

    self._entity = 'CREATIVE_ASSET'
    self.gc_project = gc_project
    self._list_name = ''
    self._id_field = FieldMap.CREATIVE_ASSET_ID
    self._search_field = None
    self.auth = auth

    self._parent_filter_name = None
    self._parent_filter_field_name = None

  def _api(self, iterate=False):
    """Returns an DCM API instance for this DAO."""
    return super(CreativeAssetDAO, self)._api(iterate).creativeAssets()

  def pre_fetch(self, feed):
    """Pre-fetches all required items to be update into the cache.

    This increases performance for update operations.

    Args:
      feed: List of feed items to retrieve
    """
    pass

  def _process_update(self, item, feed_item):
    """Handles updates to the creative asset object.

    Since creative assets are read only in DCM, there is nothing to do here,
    this method is mandatory as it is invoked by the BaseDAO class.

    Args:
      item: The creative asset DCM object being updated.
      feed_item: The feed item representing the creative asset from the
        Bulkdozer feed.
    """
    pass

  def _insert(self, new_item, feed_item):
    """Handles the upload of creative assets to DCM and the creation of the associated entity.

    This method makes a call to the DCM API to create a new entity.

    Args:
      new_item: The item to insert into DCM.
      feed_item: The feed item representing the creative asset from the
        Bulkdozer feed.

    Returns:
      The newly created item in DCM.
    """
    filename = feed_item.get(FieldMap.CREATIVE_ASSET_FILE_NAME, None)

    file_buffer = object_get(
        'user', '%s:%s' %
        (feed_item.get(FieldMap.CREATIVE_ASSET_BUCKET_NAME, None), filename))

    file_mime = mimetypes.guess_type(filename, strict=False)[0]

    media = MediaIoBaseUpload(
        BytesIO(file_buffer),
        mimetype=file_mime,
        chunksize=CHUNKSIZE,
        resumable=True)

    result = self._api().insert(
        profileId=self.profile_id,
        advertiserId=feed_item.get(FieldMap.ADVERTISER_ID, None),
        media_body=media,
        body=new_item).execute()

    return result

  def _get(self, feed_item):
    """Retrieves an item from DCM or the local cache.

    Args:
      feed_item: The feed item representing the creative asset from the
        Bulkdozer feed.

    Returns:
      Instance of the DCM object either from the API or from the local cache.
    """
    result = store.get(self._entity,
                       feed_item.get(FieldMap.CREATIVE_ASSET_ID, None))

    if not result:
      result = {
          'id': feed_item.get(FieldMap.CREATIVE_ASSET_ID, None),
          'assetIdentifier': {
              'name': feed_item.get(FieldMap.CREATIVE_ASSET_NAME, None),
              'type': feed_item.get(FieldMap.CREATIVE_TYPE, None)
          }
      }

      store.set(self._entity, [feed_item.get(FieldMap.CREATIVE_ASSET_ID, None)],
                result)

    return result

  def _update(self, item, feed_item):
    """Performs an update in DCM.

    Since this method is not allowed for creative assets because those cannot be
    updated, this method reimplements _update from BaseDAO but doesn't do
    anything to prevent an error.

    Args:
      item: The item to update in DCM.
      feed_item: The feed item representing the creative asset in the Bulkdozer
        feed.
    """
    pass

  def _process_new(self, feed_item):
    """Creates a new creative asset DCM object from a feed item representing a creative asset from the Bulkdozer feed.

    This function simply creates the object to be inserted later by the BaseDAO
    object.

    Args:
      feed_item: Feed item representing the creative asset from the Bulkdozer
        feed.

    Returns:
      A creative asset object ready to be inserted in DCM through the API.

    """
    return {
        'assetIdentifier': {
            'name': feed_item.get(FieldMap.CREATIVE_ASSET_FILE_NAME, None),
            'type': feed_item.get(FieldMap.CREATIVE_TYPE, None)
        }
    }

  def _post_process(self, feed_item, item):
    """Maps ids and names of related entities so they can be updated in the Bulkdozer feed.

    When Bulkdozer is done processing an item, it writes back the updated names
    and ids of related objects, this method makes sure those are updated in the
    creative asset feed.

    Args:
      feed_item: Feed item representing the creative asset from the Bulkdozer
        feed.
      item: The DCM creative asset being updated or created.
    """
    if item['assetIdentifier']['name']:
      feed_item[FieldMap.CREATIVE_ASSET_NAME] = item['assetIdentifier']['name']

  def get_identifier(self, association, feed):
    asset_ids = (association.get(FieldMap.CREATIVE_ASSET_ID, None),
                 store.translate(self._entity,
                                 association[FieldMap.CREATIVE_ASSET_ID]))

    for creative_asset in feed.feed:
      if creative_asset[FieldMap.CREATIVE_ASSET_ID] in asset_ids or str(
          creative_asset[FieldMap.CREATIVE_ASSET_ID]) in asset_ids:
        return {
            'name': creative_asset.get(FieldMap.CREATIVE_ASSET_NAME, None),
            'type': creative_asset.get(FieldMap.CREATIVE_TYPE, None)
        }

    return None

  def get_backup_identifier(self, association, feed):

    asset_ids = (association.get(FieldMap.CREATIVE_BACKUP_ASSET_ID, None),
                 store.translate(
                     self._entity,
                     association[FieldMap.CREATIVE_BACKUP_ASSET_ID]))

    for creative_asset in feed.feed:
      if creative_asset[FieldMap.CREATIVE_ASSET_ID] in asset_ids or str(
          creative_asset[FieldMap.CREATIVE_ASSET_ID]) in asset_ids:
        return {
            'name': creative_asset.get(FieldMap.CREATIVE_ASSET_NAME, None),
            'type': creative_asset.get(FieldMap.CREATIVE_TYPE, None)
        }

    return None



###########################################################################
#
#  Copyright 2020 Google LLC
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      https://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
#
###########################################################################

from starthinker.util.bigquery import table_create
from starthinker.util.data import get_rows
from starthinker.util.data import put_rows
from starthinker.util.google_api import API_DV360
from starthinker.util.google_api.discovery_to_bigquery import Discovery_To_BigQuery
from starthinker.util.project import project
from starthinker.util.regexp import lookup_id
from starthinker.util.sheets import sheets_clear


def creative_clear():
  table_create(
    project.task["auth_bigquery"],
    project.id,
    project.task["dataset"],
    "DV_Creatives",
    Discovery_To_BigQuery(
      "displayvideo",
      "v1"
    ).method_schema("advertisers.creatives.list"),
  )

  sheets_clear(project.task["auth_sheets"], project.task["sheet"], "Creatives", "B2:Z")


def creative_load():

  # load multiple partners from user defined sheet
  def creative_load_multiple():
    rows = get_rows(
      project.task["auth_sheets"], {
        "sheets": {
          "sheet": project.task["sheet"],
          "tab": "Advertisers",
          "range": "A2:A"
        }
    })

    for row in rows:
      yield from API_DV360(
        project.task["auth_dv"], iterate=True).advertisers().creatives().list(
          advertiserId=lookup_id(row[0])
        ).execute()

  # write creatives to database and sheet
  put_rows(
    project.task["auth_bigquery"], {
      "bigquery": {
        "dataset":project.task["dataset"],
        "table":"DV_Creatives",
        "schema":Discovery_To_BigQuery(
          "displayvideo",
          "v1"
        ).method_schema("advertisers.creatives.list"),
        "format": "JSON"
      }
    }, creative_load_multiple()
  )

  # write creatives to sheet
  rows = get_rows(
    project.task["auth_bigquery"], {
      "bigquery": {
        "dataset":project.task["dataset"],
        "query":"""SELECT
          CONCAT(P.displayName, ' - ', P.partnerId),
          CONCAT(A.displayName, ' - ', A.advertiserId),
          CONCAT(C.displayName, ' - ', C.creativeId),
          C.entityStatus,
          C.creativeType,
          C.dimensions.widthPixels,
          C.dimensions.heightPixels,
          C.reviewStatus.approvalStatus,
          C.reviewStatus.creativeAndLandingPageReviewStatus,
          C.reviewStatus.contentAndPolicyReviewStatus,
          COUNTIF(RS.status='REVIEW_STATUS_UNSPECIFIED') OVER() AS Exchanges_Unspecified,
          COUNTIF(RS.status='REVIEW_STATUS_PENDING') OVER() AS Exchanges_Pending,
          COUNTIF(RS.status='REVIEW_STATUS_REJECTED') OVER() AS Exchanges_Rejected,
          COUNTIF(RS.status='REVIEW_STATUS_APPROVED') OVER() AS Exchanges_Approved,
          COUNTIF(RP.status='REVIEW_STATUS_UNSPECIFIED') OVER() AS Publishers_Unspecified,
          COUNTIF(RP.status='REVIEW_STATUS_PENDING') OVER() AS Publishers_Pending,
          COUNTIF(RP.status='REVIEW_STATUS_REJECTED') OVER() AS Publishers_Rejected,
          COUNTIF(RP.status='REVIEW_STATUS_APPROVED') OVER() AS Publishers_Approved,
          FROM `{dataset}.DV_Creatives` AS C, UNNEST(reviewStatus.exchangeReviewStatuses) AS RS, UNNEST(reviewStatus.publisherReviewStatuses) AS RP
          LEFT JOIN `{dataset}.DV_Advertisers` AS A
          ON C.advertiserId=A.advertiserId
          LEFT JOIN `{dataset}.DV_Partners` AS P
          ON A.partnerId=P.partnerId
        """.format(**project.task),
        "legacy":False
      }
    }
  )

  put_rows(
    project.task["auth_sheets"],
    {
      "sheets": {
        "sheet": project.task["sheet"],
        "tab": "Creatives",
        "range": "B2"
      }
    },
    rows
  )



###########################################################################
#
#  Copyright 2020 Google LLC
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      https://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
#
###########################################################################

from starthinker.util.project import project
from starthinker.util.csv import rows_to_type
from starthinker.util.sheets import sheets_clear
from starthinker.util.sheets import sheets_write
from starthinker.util.data import get_rows, put_rows
from starthinker.util.bigquery import query_to_table, query_to_view, storage_to_table, query_to_rows, rows_to_table, run_query, query_parameters
from starthinker.util.bigquery.functions import pearson_significance_test


@project.from_parameters
def bigquery():

  if 'function' in project.task:
    query = None

    if project.task['function'] == 'pearson_significance_test':
      query = pearson_significance_test()

    if query:
      run_query(project.task['auth'], project.id, query, False,
                project.task['to']['dataset'])

  elif 'run' in project.task and 'query' in project.task.get('run', {}):
    if project.verbose:
      print('QUERY', project.task['run']['query'])
    run_query(
        project.task['auth'],
        project.id,
        query_parameters(project.task['run']['query'],
                         project.task['run'].get('parameters')),
        project.task['run'].get('legacy', True),
    )

  elif 'values' in project.task['from']:
    rows = get_rows(project.task['auth'], project.task['from'])

    rows_to_table(project.task['to'].get('auth', project.task['auth']),
                  project.id, project.task['to']['dataset'],
                  project.task['to']['table'], rows,
                  project.task.get('schema', []), 0)

  elif 'query' in project.task['from']:

    if 'table' in project.task['to']:
      if project.verbose:
        print('QUERY TO TABLE', project.task['to']['table'])

      query_to_table(
          project.task['auth'],
          project.id,
          project.task['to']['dataset'],
          project.task['to']['table'],
          query_parameters(project.task['from']['query'],
                           project.task['from'].get('parameters')),
          disposition=project.task['write_disposition']
          if 'write_disposition' in project.task else 'WRITE_TRUNCATE',
          legacy=project.task['from'].get(
              'legacy',
              project.task['from'].get('useLegacySql',
                                       True)),  # DEPRECATED: useLegacySql,
          target_project_id=project.task['to'].get('project_id', project.id))

    elif 'sheet' in project.task['to']:
      if project.verbose:
        print('QUERY TO SHEET', project.task['to']['sheet'])
      rows = query_to_rows(
          project.task['auth'],
          project.id,
          project.task['from']['dataset'],
          query_parameters(project.task['from']['query'],
                           project.task['from'].get('parameters')),
          legacy=project.task['from'].get('legacy', True))

      # makes sure types are correct in sheet
      rows = rows_to_type(rows)

      sheets_clear(project.task['to'].get('auth', project.task['auth']),
                   project.task['to']['sheet'], project.task['to']['tab'],
                   project.task['to'].get('range', 'A2'))
      sheets_write(project.task['to'].get('auth', project.task['auth']),
                   project.task['to']['sheet'], project.task['to']['tab'],
                   project.task['to'].get('range', 'A2'), rows)

    elif 'sftp' in project.task['to']:
      rows = query_to_rows(
          project.task['auth'],
          project.id,
          project.task['from']['dataset'],
          query_parameters(project.task['from']['query'],
                           project.task['from'].get('parameters')),
          legacy=project.task['from'].get('use_legacy_sql', True))

      if rows:
        if project.verbose:
          print('QUERY TO SFTP')
        put_rows(project.task['auth'], project.task['to'], rows)

    else:
      if project.verbose:
        print('QUERY TO VIEW', project.task['to']['view'])
      query_to_view(
          project.task['auth'],
          project.id,
          project.task['to']['dataset'],
          project.task['to']['view'],
          query_parameters(project.task['from']['query'],
                           project.task['from'].get('parameters')),
          project.task['from'].get(
              'legacy',
              project.task['from'].get('useLegacySql',
                                       True)),  # DEPRECATED: useLegacySql
          project.task['to'].get('replace', False))

  else:
    if project.verbose:
      print('STORAGE TO TABLE', project.task['to']['table'])
    storage_to_table(
        project.task['auth'], project.id, project.task['to']['dataset'],
        project.task['to']['table'],
        project.task['from']['bucket'] + ':' + project.task['from']['path'],
        project.task.get('schema', []), project.task.get('skip_rows', 1),
        project.task.get('structure', 'CSV'),
        project.task.get('disposition', 'WRITE_TRUNCATE'))


if __name__ == '__main__':
  bigquery()



from starthinker.util.project import project
from starthinker.util.trix import trix_read, trix_update
from starthinker.util.bigquery import query_to_rows

TEMPLATE_QUERY = """
SELECT
  COUNT(DISTINCT User_ID) reach,
  STRFTIME_UTC_USEC(INTEGER(Event_Time), '%%Y%%m') month,
  COUNT(DISTINCT DBM_Insertion_Order_ID) insertion_orders,
  SUM(FLOAT(DBM_Media_Cost__USD_) / 1000000000) spend,
  SUM(FLOAT(DBM_Revenue__USD_) / 1000000000) revenue
FROM
  [%s.DataTransfer_impression]
WHERE
  DBM_Advertiser_ID IN (%s)
  AND STRFTIME_UTC_USEC(INTEGER(Event_Time), '%%Y%%m%%d') BETWEEN '%s' AND '%s'
  AND REGEXP_MATCH(DBM_Matching_Targeted_Segments, r'(^| )(%s)( |$)')
GROUP BY
  month
ORDER BY
  month asc
"""


@project.from_parameters
def audience_reach_report():
  if project.verbose:
    print('AUDIENCE REACH REPORT')

  configs = trix_read(project.task['auth'], project.task['trix_id'],
                      project.task['config_range'])
  dataset = project.task['dataset']
  advertiser_ids = []
  start_date = ''
  end_date = ''
  segments = []

  for config in configs['values']:
    if config[0]:
      advertiser_ids.append('\'' + config[0] + '\'')

    if not start_date and config[1]:
      start_date = config[1]

    if not end_date and config[2]:
      end_date = config[2]

    if config[3]:
      segments.append(config[3])

  sql = TEMPLATE_QUERY % (dataset, ','.join(advertiser_ids), start_date,
                          end_date, '|'.join(segments))

  #result = run_query(sql, dataset_name=dataset)
  #trix_update(project.task['auth'], project.task['trix_id'], project.task['output_range'], {'values': result[0]}, clear=True)

  rows = query_to_rows(project.task['auth'], project.id, dataset, sql)
  trix_update(
      project.task['auth'],
      project.task['trix_id'],
      project.task['output_range'], {'values': rows},
      clear=True)


if __name__ == '__main__':
  audience_reach_report()



###########################################################################
#
#  Copyright 2020 Google LLC
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      https://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
#
###########################################################################
"""A set of operators to facilitate StarThinker integration with AirFlow.

The operators are meant to facilitate the reuse of StarThinker code inside an
AirFlow workflow, for example being able to reuse a StarThinker task without
having to specify the entire recipe.
"""

import importlib
import json

from airflow.hooks.base_hook import BaseHook
from airflow.operators.python_operator import PythonOperator
from airflow.utils.decorators import apply_defaults


class StarThinkerTaskOperator(PythonOperator):
  """Runs a StarThinker task.

  This operator takes care of the creation of a valid StarThinker recipe JSON
  containing a single task and its parameters, as well as valid authentication
  information.

  Attributes:
    starthinker_task: A string identifying the StarThinker task to be called.
    starthinker_task_config: A dictionary containing the StarThinker task
      configuration. (templated) Can receive a json string representing the task
      configuration or reference to a template file. Template references are
      recognized by a string ending in '.json'.
    user_conn_id: The connection to use for user authentication.
    gcp_conn_id: The connection to use for service authentication.
    delegate_to: The account to impersonate, if any. Currently ignored.
  """

  template_fields = ['starthinker_task_config']
  template_ext = ['.json']
  ui_color = '#C4C4C4'

  @apply_defaults
  def __init__(self,
               starthinker_task,
               starthinker_task_config,
               user_conn_id=None,
               gcp_conn_id='google_cloud_default',
               delegate_to=None,
               *args,
               **kwargs):
    # Generate the correct callable for the task, based on the assumption that
    # StarThinker tasks are located under plugins/starthinker/task
    starthinker_callable = getattr(
        importlib.import_module('starthinker.task.%s.run' % starthinker_task),
        starthinker_task)

    super(StarThinkerTaskOperator, self).__init__(
        python_callable=starthinker_callable, *args, **kwargs)
    self.starthinker_task = starthinker_task
    self.starthinker_task_config = starthinker_task_config
    self.user_conn_id = user_conn_id
    self.gcp_conn_id = gcp_conn_id
    self.delegate_to = delegate_to

    # A baseline recipe that will have only one task
    self.recipe = {
        'setup': {
            'auth': {
                'user': '',
                'service': ''
            }
        },
        'tasks': [{
            starthinker_task: starthinker_task_config
        }]
    }

    # If supplied, load "user" auth information into the recipe
    if user_conn_id:
      user_connection_extra = BaseHook.get_connection(user_conn_id).extra_dejson
      if user_connection_extra['extra__google_cloud_platform__key_path']:
        self.recipe['setup']['auth']['user'] = user_connection_extra[
            'extra__google_cloud_platform__key_path']
      elif user_connection_extra['extra__google_cloud_platform__keyfile_dict']:
        self.recipe['setup']['auth']['user'] = user_connection_extra[
            'extra__google_cloud_platform__keyfile_dict']
      if user_connection_extra['extra__google_cloud_platform__project']:
        self.recipe['setup']['id'] = user_connection_extra[
            'extra__google_cloud_platform__project']

    # Load "service" auth information into the recipe
    if gcp_conn_id:
      service_connection_extra = BaseHook.get_connection(
          gcp_conn_id).extra_dejson
      if service_connection_extra['extra__google_cloud_platform__key_path']:
        self.recipe['setup']['auth']['service'] = service_connection_extra[
            'extra__google_cloud_platform__key_path']
      elif service_connection_extra[
          'extra__google_cloud_platform__keyfile_dict']:
        self.recipe['setup']['auth']['service'] = service_connection_extra[
            'extra__google_cloud_platform__keyfile_dict']
        try:
          keyfile_dict_json = json.loads(service_connection_extra[
              'extra__google_cloud_platform__keyfile_dict'])
          if keyfile_dict_json and keyfile_dict_json['project_id']:
            self.recipe['setup']['id'] = keyfile_dict_json['project_id']
        except Exception as e:
          self.log.error('Failed parsing the Keyfile JSON for gcp_conn_id=%s',
                         self.gcp_conn_id)
          raise e
      if service_connection_extra['extra__google_cloud_platform__project']:
        self.recipe['setup']['id'] = service_connection_extra[
            'extra__google_cloud_platform__project']

    self.log.info('The compiled recipe is the following:\n\n%s' %
                  json.dumps(self.recipe))

    self.op_kwargs = {'recipe': self.recipe, 'instance': 1}



###########################################################################
#
#  Copyright 2020 Google LLC
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      https://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
#
###########################################################################

from starthinker.util.bigquery import table_create
from starthinker.util.data import get_rows
from starthinker.util.data import put_rows
from starthinker.util.google_api import API_DV360
from starthinker.util.google_api.discovery_to_bigquery import Discovery_To_BigQuery
from starthinker.util.project import project
from starthinker.util.regexp import lookup_id


def insertion_order_clear():
  table_create(
    project.task["auth_bigquery"],
    project.id,
    project.task["dataset"],
    "DV_InsertionOrders",
    Discovery_To_BigQuery(
      "displayvideo",
      "v1"
    ).method_schema(
      "advertisers.insertionOrders.list"
    )
  )

def insertion_order_load():

  # load multiple from user defined sheet
  def insertion_order_load_multiple():
    rows = get_rows(
      project.task["auth_sheets"],
      { "sheets": {
        "sheet": project.task["sheet"],
        "tab": "Advertisers",
        "range": "A2:A"
      }}
    )

    for row in rows:
      yield from API_DV360(
        project.task["auth_dv"],
        iterate=True
      ).advertisers().insertionOrders().list(
        advertiserId=lookup_id(row[0])
      ).execute()

  # write insertion orders to database
  put_rows(
    project.task["auth_bigquery"],
    { "bigquery": {
      "dataset": project.task["dataset"],
      "table": "DV_InsertionOrders",
      "schema": Discovery_To_BigQuery(
        "displayvideo",
        "v1"
      ).method_schema("advertisers.insertionOrders.list"),
      "format": "JSON"
    }},
    insertion_order_load_multiple()
  )



###########################################################################
#
#  Copyright 2020 Google LLC
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      https://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
#
###########################################################################

import os
import re
import json
import pprint
from copy import deepcopy
from datetime import datetime, date

from django.conf import settings

from starthinker.config import UI_ROOT
from starthinker.script.parse import json_set_fields, text_set_fields
from starthinker.util.project import get_project

# cache scripts in memory
SCRIPTS = {}
APACHE_LICENSE_MATCH = 'Licensed under the Apache License, Version 2.0'


def load_scripts():
  if not SCRIPTS:
    for root, dirs, files in os.walk(UI_ROOT + '/scripts/'):
      for filename in sorted(files):
        if filename.endswith('.json'):
          try:
            script = get_project(root + filename)
            if not 'script' in script:
              continue
            script['path'] = root + filename
            SCRIPTS[filename.replace('.json', '', 1)] = script
            print('OK', filename)
          except Exception as e:
            print('ERROR:', filename, str(e))


# Try to load from the lookup script ( appengine can't read files so need to generate script )
try:
  from starthinker_ui.recipe.scripts_lookup import SCRIPTS
  print('Loading From: starthinker.starthinker_ui.recipe.scripts_lookup')
  print(
      'To load from scripts directly delete: starthinker/starthinker_ui/recipe/scripts_lookup.py'
  )
# If it fails, load the scripts directly from the json files
except ImportError as e:
  print('Loading From: scripts/*.json')
  load_scripts()


class Script:

  @staticmethod
  def get_scripts(account_email=None, ui=False):
    for tag in sorted(iter(SCRIPTS.keys())):
      if account_email in SCRIPTS[tag]['script'].get('private',
                                                     (account_email,)):
        if not ui or ('from' in SCRIPTS[tag]['script'] and
                      'to' in SCRIPTS[tag]['script']):
          yield Script(tag)

  def __init__(self, tag):
    self.tag = tag
    self.script = SCRIPTS.get(tag, {})

  def exists(self):
    return self.script != {}

  def get_tag(self):
    return self.tag

  def get_link(self):
    return '%s/solution/%s/' % (
        settings.CONST_URL,
        self.tag,
    )

  def get_link_client(self):
    if self.script.get('script', {}).get('license', '') == APACHE_LICENSE_MATCH:
      return 'https://google.github.io/starthinker/solution/%s/' % self.get_tag()
    else:
      return ''

  def get_link_ui(self):
    return '/recipe/%s/?script=%s' % ('manual' if self.is_manual() else 'edit', self.get_tag())

  def get_link_colab(self):
    return 'https://colab.research.google.com/github/google/starthinker/blob/master/colabs/%s.ipynb' % self.get_tag()

  def get_link_airflow(self):
    return 'https://github.com/google/starthinker/blob/master/dags/%s_dag.py' % self.get_tag()

  def get_open_source(self):
    if self.script.get('script', {}).get('license', '') == APACHE_LICENSE_MATCH:
      return 'https://github.com/google/starthinker/blob/master/scripts/%s.json' % self.get_tag()
    else:
      return ''

  def get_released(self):
    try:
      return datetime.strptime(
          self.script.get('script', {}).get('released', ''), '%Y-%m-%d').date()
    except:
      return None

  def get_released_ago(self):
    agos = []
    days = (date.today() - self.get_released()).days

    if days <= 7:
      agos.append('7 days')
    if days <= 30:
      agos.append('30 days')
    if days <= 90:
      agos.append('90 days')
    if days > 90:
      agos.append('older')

    return agos

  def get_name(self):
    return self.script.get('script', {}).get('title', '')

  def get_icon(self):
    return self.script.get('script', {}).get('icon', '')

  def get_description(self, variables={}):
    return text_set_fields(
        self.script.get('script', {}).get('description', ''), variables)

  def get_instructions(self, variables={}):
    return [
        text_set_fields(instruction, variables)
        for instruction in self.script.get('script', {}).get(
            'instructions', [])
    ]

  def get_authors(self):
    return set(deepcopy(self.script.get('script', {}).get('authors', [])))

  def get_image(self):
    return self.script.get('script', {}).get('image', None)

  def get_sample(self):
    return self.script.get('script', {}).get('sample', None)

  def get_test(self):
    return self.script.get('script', {}).get('test', None)

  def get_documentation(self):
    return self.script.get('script', {}).get('documentation', None)

  def get_from(self):
    return self.script.get('script', {}).get('from', [])

  def get_to(self):
    return self.script.get('script', {}).get('to', [])

  def get_pitches(self):
    return self.script.get('script', {}).get('pitches', [])

  def get_impacts(self):
    return self.script.get('script', {}).get('impacts', {})

  def get_tasks(self):
    return deepcopy(self.script.get('tasks', []))

  def get_tasks_and_setup(self):
    s = deepcopy(self.script)
    if 'script' in s: del s['script']
    if 'path' in s: del s['path']
    return s

  def get_tasks_linked(self):
    tasks = self.get_tasks()
    data = json.dumps(tasks, indent=4)
    for task in set([next(iter(task.keys())) for task in tasks]):
      data = re.sub(
          r'\n {8}"%s": {' % task,
          '\n        "<a href="https://github.com/google/starthinker/tree/master/starthinker/task/%s/run.py" target="_blank">%s</a>": {'
          % (task, task), data)
    return data

  def is_manual(self):
    return self.script.get('setup', {}).get('day', None) == []

  @staticmethod
  def get_json(
    uuid,
    project_id,
    project_key,
    credentials_user,
    credentials_service,
    timezone,
    days,
    hours,
    values
    ):
    tasks = []
    hours = set(hours)
    for v in values:
      ts = Script(v['tag']).get_tasks()
      json_set_fields(ts, v['values'])
      tasks.extend(ts)

    data = {
        #"script":{
        #  "tag":self.get_tag(),
        #  "authors":self.get_authors()
        #},
        'setup': {
            'uuid': uuid,
            'id': project_id,
            'key': project_key,
            'timezone': timezone,
            'day': days,
            'hour': sorted(hours),
            'auth': {
                'user': credentials_user,
                'service': credentials_service,
            }
        },
        'tasks': tasks,
    }

    return data


if __name__ == '__main__':
  """ Create a scripts_lookup.py file to load from instead of json scripts, used by appengine.

  The scripts_lookup.py file acts as a buffer storing all the configurations for
  all the scripts.
  It is necessary because some servers like Google App Engine cannot build the
  lookup using the file system.
  For development you can delete: scripts_lookup.py, the script templats will be
  loaded directly from files.

  """

  filename = '%s/starthinker_ui/recipe/scripts_lookup.py' % UI_ROOT
  print('WRITING:', filename)
  content = ('# DO NOT EDIT, GENERATED BY: python '
             'starthinker_ui/recipe/scripts.py\n\nSCRIPTS = %s') % pprint.pformat(
      SCRIPTS, indent=1)
  with open(filename, 'w') as file:
    file.write(content)



###########################################################################
#
#  Copyright 2020 Google LLC
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      https://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
#
###########################################################################

Advertiser_Schema = [
    {
        'name':
            'common_data',
        'type':
            'RECORD',
        'mode':
            'NULLABLE',
        'fields': [
            {
                'name': 'id',
                'type': 'INTEGER',
                'mode': 'NULLABLE',
            },
            {
                'name': 'name',
                'type': 'STRING',
                'mode': 'NULLABLE',
            },
            {
                'name': 'active',
                'type': 'BOOLEAN',
                'mode': 'NULLABLE',
            },
            {
                'name': 'integration_code',
                'type': 'STRING',
                'mode': 'NULLABLE',
            },
        ]
    },
    {
        'name': 'partner_id',
        'type': 'INTEGER',
        'mode': 'NULLABLE',
    },
    {
        'name': 'currency_code',
        'type': 'STRING',
        'mode': 'NULLABLE',
    },
    {
        'name': 'timezone_code',
        'type': 'STRING',
        'mode': 'NULLABLE',
    },
    {
        'name': 'landing_page_url',
        'type': 'STRING',
        'mode': 'NULLABLE',
    },
    {
        'name': 'available_channel_ids',
        'type': 'INTEGER',
        'mode': 'REPEATED',
    },
    {
        'name': 'blacklist_channel_id',
        'type': 'INTEGER',
        'mode': 'REPEATED',
    },
    {
        'name': 'dcm_configuration',
        'type': 'INTEGER',
        'mode': 'NULLABLE',
    },
    {
        'name': 'dcm_network_id',
        'type': 'INTEGER',
        'mode': 'NULLABLE',
    },
    {
        'name': 'dcm_network_name',
        'type': 'STRING',
        'mode': 'NULLABLE',
    },
    {
        'name': 'dcm_advertiser_id',
        'type': 'INTEGER',
        'mode': 'REPEATED',
    },
    {
        'name': 'dcm_advertiser_name',
        'type': 'STRING',
        'mode': 'NULLABLE',
    },
    {
        'name': 'dcm_floodlight_group_name',
        'type': 'STRING',
        'mode': 'NULLABLE',
    },
    {
        'name': 'dcm_syncable_site_ids',
        'type': 'INTEGER',
        'mode': 'REPEATED',
    },
    {
        'name': 'enable_oba_tags',
        'type': 'BOOLEAN',
        'mode': 'NULLABLE',
    },
]



###########################################################################
#
#  Copyright 2020 Google LLC
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      https://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
#
###########################################################################

PartnerCosts_Schema = [
    {
        'name': 'cpm_fee_1_advertiser_micros',
        'type': 'INTEGER',
        'mode': 'NULLABLE',
    },
    {
        'name': 'cpm_fee_2_advertiser_micros',
        'type': 'INTEGER',
        'mode': 'NULLABLE',
    },
    {
        'name': 'cpm_fee_3_advertiser_micros',
        'type': 'INTEGER',
        'mode': 'NULLABLE',
    },
    {
        'name': 'cpm_fee_4_advertiser_micros',
        'type': 'INTEGER',
        'mode': 'NULLABLE',
    },
    {
        'name': 'cpm_fee_5_advertiser_micros',
        'type': 'INTEGER',
        'mode': 'NULLABLE',
    },
    {
        'name': 'media_fee_percent_1_millis',
        'type': 'INTEGER',
        'mode': 'NULLABLE',
    },
    {
        'name': 'media_fee_percent_2_millis',
        'type': 'INTEGER',
        'mode': 'NULLABLE',
    },
    {
        'name': 'media_fee_percent_3_millis',
        'type': 'INTEGER',
        'mode': 'NULLABLE',
    },
    {
        'name': 'media_fee_percent_4_millis',
        'type': 'INTEGER',
        'mode': 'NULLABLE',
    },
    {
        'name': 'media_fee_percent_5_millis',
        'type': 'INTEGER',
        'mode': 'NULLABLE',
    },
    {
        'name': 'cpm_fee_1_cost_type',
        'type': 'INTEGER',
        'mode': 'NULLABLE',
    },
    {
        'name': 'cpm_fee_2_cost_type',
        'type': 'INTEGER',
        'mode': 'NULLABLE',
    },
    {
        'name': 'cpm_fee_3_cost_type',
        'type': 'INTEGER',
        'mode': 'NULLABLE',
    },
    {
        'name': 'cpm_fee_4_cost_type',
        'type': 'INTEGER',
        'mode': 'NULLABLE',
    },
    {
        'name': 'cpm_fee_5_cost_type',
        'type': 'INTEGER',
        'mode': 'NULLABLE',
    },
    {
        'name': 'media_fee_percent_1_cost_type',
        'type': 'INTEGER',
        'mode': 'NULLABLE',
    },
    {
        'name': 'media_fee_percent_2_cost_type',
        'type': 'INTEGER',
        'mode': 'NULLABLE',
    },
    {
        'name': 'media_fee_percent_3_cost_type',
        'type': 'INTEGER',
        'mode': 'NULLABLE',
    },
    {
        'name': 'media_fee_percent_4_cost_type',
        'type': 'INTEGER',
        'mode': 'NULLABLE',
    },
    {
        'name': 'media_fee_percent_5_cost_type',
        'type': 'INTEGER',
        'mode': 'NULLABLE',
    },
    {
        'name': 'cpm_fee_1_bill_to_type',
        'type': 'INTEGER',
        'mode': 'NULLABLE',
    },
    {
        'name': 'cpm_fee_2_bill_to_type',
        'type': 'INTEGER',
        'mode': 'NULLABLE',
    },
    {
        'name': 'cpm_fee_3_bill_to_type',
        'type': 'INTEGER',
        'mode': 'NULLABLE',
    },
    {
        'name': 'cpm_fee_4_bill_to_type',
        'type': 'INTEGER',
        'mode': 'NULLABLE',
    },
    {
        'name': 'cpm_fee_5_bill_to_type',
        'type': 'INTEGER',
        'mode': 'NULLABLE',
    },
    {
        'name': 'media_fee_percent_1_bill_to_type',
        'type': 'INTEGER',
        'mode': 'NULLABLE',
    },
    {
        'name': 'media_fee_percent_2_bill_to_type',
        'type': 'INTEGER',
        'mode': 'NULLABLE',
    },
    {
        'name': 'media_fee_percent_3_bill_to_type',
        'type': 'INTEGER',
        'mode': 'NULLABLE',
    },
    {
        'name': 'media_fee_percent_4_bill_to_type',
        'type': 'INTEGER',
        'mode': 'NULLABLE',
    },
    {
        'name': 'media_fee_percent_5_bill_to_type',
        'type': 'INTEGER',
        'mode': 'NULLABLE',
    },
]



###########################################################################
#
#  Copyright 2020 Google LLC
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      https://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
#
###########################################################################

from starthinker.util.bigquery import table_create
from starthinker.util.data import get_rows
from starthinker.util.data import put_rows
from starthinker.util.google_api import API_DV360
from starthinker.util.google_api.discovery_to_bigquery import Discovery_To_BigQuery
from starthinker.util.project import project
from starthinker.util.regexp import lookup_id


def campaign_clear():
  table_create(
    project.task['auth_bigquery'],
    project.id,
    project.task['dataset'],
    'DV_Campaigns',
    Discovery_To_BigQuery(
      'displayvideo',
      'v1'
    ).method_schema(
      'advertisers.campaigns.list'
    )
  )


def campaign_load():

  # load multiple partners from user defined sheet
  def campaign_load_multiple():
    rows = get_rows(
      project.task['auth_sheets'], 
      { 'sheets': {
        'sheet': project.task['sheet'],
        'tab': 'Advertisers',
        'range': 'A2:A'
      }}
    )

    for row in rows:
      yield from API_DV360(
        project.task['auth_dv'],
        iterate=True
      ).advertisers().campaigns().list(
        advertiserId=lookup_id(row[0])
      ).execute()

  # write campaigns to database
  put_rows(
    project.task['auth_bigquery'], 
    { 'bigquery': {
      'dataset': project.task['dataset'],
      'table': 'DV_Campaigns',
      'schema': Discovery_To_BigQuery(
        'displayvideo',
        'v1'
      ).method_schema(
        'advertisers.campaigns.list'
      ),
      'format': 'JSON'
    }},
    campaign_load_multiple()
  )



###########################################################################
#
#  Copyright 2020 Google LLC
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      https://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
#
###########################################################################

'''
--------------------------------------------------------------

Before running this Airflow module...

  Install StarThinker in cloud composer ( recommended ):

    From Release: pip install starthinker
    From Open Source: pip install git+https://github.com/google/starthinker

  Or push local code to the cloud composer plugins directory ( if pushing local code changes ):

    source install/deploy.sh
    4) Composer Menu
    l) Install All

--------------------------------------------------------------

  If any recipe task has "auth" set to "user" add user credentials:

    1. Ensure an RECIPE['setup']['auth']['user'] = [User Credentials JSON]

  OR

    1. Visit Airflow UI > Admin > Connections.
    2. Add an Entry called "starthinker_user", fill in the following fields. Last step paste JSON from authentication.
      - Conn Type: Google Cloud Platform
      - Project: Get from https://github.com/google/starthinker/blob/master/tutorials/cloud_project.md
      - Keyfile JSON: Get from: https://github.com/google/starthinker/blob/master/tutorials/deploy_commandline.md#optional-setup-user-credentials

--------------------------------------------------------------

  If any recipe task has "auth" set to "service" add service credentials:

    1. Ensure an RECIPE['setup']['auth']['service'] = [Service Credentials JSON]

  OR

    1. Visit Airflow UI > Admin > Connections.
    2. Add an Entry called "starthinker_service", fill in the following fields. Last step paste JSON from authentication.
      - Conn Type: Google Cloud Platform
      - Project: Get from https://github.com/google/starthinker/blob/master/tutorials/cloud_project.md
      - Keyfile JSON: Get from: https://github.com/google/starthinker/blob/master/tutorials/cloud_service.md

--------------------------------------------------------------

DV360 User Audit

Gives DV clients ability to see which users have access to which parts of an account. Loads DV user profile mappings using the API into BigQuery and connects to a DataStudio dashboard.

  - DV360 only permits SERVICE accounts to access the user list API endpoint, be sure to provide and permission one.
  - Wait for <b>BigQuery->UNDEFINED->UNDEFINED->DV_*</b> to be created.
  - Wait for <b>BigQuery->UNDEFINED->UNDEFINED->Barnacle_*</b> to be created, then copy and connect the following data sources.
  - Join the <a hre='https://groups.google.com/d/forum/starthinker-assets' target='_blank'>StarThinker Assets Group</a> to access the following assets
  - Copy <a href='https://datastudio.google.com/c/u/0/reporting/9f6b9e62-43ec-4027-849a-287e9c1911bd' target='_blank'>Barnacle DV Report</a>.
  - Click Edit->Resource->Manage added data sources, then edit each connection to connect to your new tables above.
  - Or give these intructions to the client.

--------------------------------------------------------------

This StarThinker DAG can be extended with any additional tasks from the following sources:
  - https://google.github.io/starthinker/
  - https://github.com/google/starthinker/tree/master/dags

'''

from starthinker.airflow.factory import DAG_Factory

INPUTS = {
  'auth_read': 'user',  # Credentials used for writing data.
  'auth_write': 'service',  # Credentials used for writing data.
  'partner': '',  # Partner ID to run user audit on.
  'recipe_slug': '',  # Name of Google BigQuery dataset to create.
}

RECIPE = {
  'setup': {
    'day': [
      'Mon',
      'Tue',
      'Wed',
      'Thu',
      'Fri',
      'Sat',
      'Sun'
    ],
    'hour': [
      3,
      15
    ]
  },
  'tasks': [
    {
      'dataset': {
        'auth': {
          'field': {
            'name': 'auth_write',
            'kind': 'authentication',
            'order': 1,
            'default': 'service',
            'description': 'Credentials used for writing data.'
          }
        },
        'dataset': {
          'field': {
            'name': 'recipe_slug',
            'kind': 'string',
            'order': 4,
            'default': '',
            'description': 'Name of Google BigQuery dataset to create.'
          }
        }
      }
    },
    {
      'google_api': {
        'auth': {
          'field': {
            'name': 'auth_read',
            'kind': 'authentication',
            'order': 0,
            'default': 'user',
            'description': 'Credentials used for writing data.'
          }
        },
        'api': 'doubleclickbidmanager',
        'version': 'v1.1',
        'function': 'queries.listqueries',
        'alias': 'list',
        'results': {
          'bigquery': {
            'auth': {
              'field': {
                'name': 'auth_write',
                'kind': 'authentication',
                'order': 1,
                'default': 'service',
                'description': 'Credentials used for writing data.'
              }
            },
            'dataset': {
              'field': {
                'name': 'recipe_slug',
                'kind': 'string',
                'order': 4,
                'default': '',
                'description': 'Name of Google BigQuery dataset to create.'
              }
            },
            'table': 'DV_Reports'
          }
        }
      }
    },
    {
      'google_api': {
        'auth': {
          'field': {
            'name': 'auth_read',
            'kind': 'authentication',
            'order': 0,
            'default': 'user',
            'description': 'Credentials used for writing data.'
          }
        },
        'api': 'displayvideo',
        'version': 'v1',
        'function': 'partners.list',
        'kwargs': {'fields': 'partners.displayName,partners.partnerId,nextPageToken'},
        'results': {
          'bigquery': {
            'auth': {
              'field': {
                'name': 'auth_write',
                'kind': 'authentication',
                'order': 1,
                'default': 'service',
                'description': 'Credentials used for writing data.'
              }
            },
            'dataset': {
              'field': {
                'name': 'recipe_slug',
                'kind': 'string',
                'order': 4,
                'default': '',
                'description': 'Name of Google BigQuery dataset to create.'
              }
            },
            'table': 'DV_Partners'
          }
        }
      }
    },
    {
      'google_api': {
        'auth': {
          'field': {
            'name': 'auth_read',
            'kind': 'authentication',
            'order': 0,
            'default': 'user',
            'description': 'Credentials used for writing data.'
          }
        },
        'api': 'displayvideo',
        'version': 'v1',
        'function': 'advertisers.list',
        'kwargs': {'partnerId': {'field': {'name': 'partner','kind': 'integer','order': 2,'default': '','description': 'Partner ID to run user audit on.'}},'fields': 'advertisers.displayName,advertisers.advertiserId,nextPageToken'},
        'results': {
          'bigquery': {
            'auth': {
              'field': {
                'name': 'auth_write',
                'kind': 'authentication',
                'order': 1,
                'default': 'service',
                'description': 'Credentials used for writing data.'
              }
            },
            'dataset': {
              'field': {
                'name': 'recipe_slug',
                'kind': 'string',
                'order': 4,
                'default': '',
                'description': 'Name of Google BigQuery dataset to create.'
              }
            },
            'table': 'DV_Advertisers'
          }
        }
      }
    },
    {
      'google_api': {
        'auth': 'service',
        'api': 'displayvideo',
        'version': 'v1',
        'function': 'users.list',
        'kwargs': {
        },
        'results': {
          'bigquery': {
            'auth': {
              'field': {
                'name': 'auth_write',
                'kind': 'authentication',
                'order': 1,
                'default': 'service',
                'description': 'Credentials used for writing data.'
              }
            },
            'dataset': {
              'field': {
                'name': 'recipe_slug',
                'kind': 'string',
                'order': 4,
                'default': '',
                'description': 'Name of Google BigQuery dataset to create.'
              }
            },
            'table': 'DV_Users'
          }
        }
      }
    },
    {
      'bigquery': {
        'auth': {
          'field': {
            'name': 'auth_write',
            'kind': 'authentication',
            'order': 1,
            'default': 'service',
            'description': 'Credentials used for writing data.'
          }
        },
        'from': {
          'query': "SELECT           U.userId,           U.name,           U.email,           U.displayName,           REGEXP_EXTRACT(U.email, r'@(.+)') AS Domain,           IF (ENDS_WITH(U.email, '.gserviceaccount.com'), 'Service', 'User') AS Authentication,           IF((Select COUNT(advertiserId) from UNNEST(U.assignedUserRoles)) = 0, 'Partner', 'Advertiser') AS Scope,           STRUCT(             AUR.partnerId,             P.displayName AS partnerName,             AUR.userRole,             AUR.advertiserId,             A.displayName AS advertiserName,             AUR.assignedUserRoleId           ) AS assignedUserRoles,           FROM `{dataset}.DV_Users` AS U,           UNNEST(assignedUserRoles) AS AUR           LEFT JOIN `{dataset}.DV_Partners` AS P           ON AUR.partnerId=P.partnerId           LEFT JOIN `{dataset}.DV_Advertisers` AS A           ON AUR.advertiserId=A.advertiserId         ",
          'parameters': {
            'dataset': {
              'field': {
                'name': 'recipe_slug',
                'kind': 'string',
                'order': 4,
                'default': '',
                'description': 'Name of Google BigQuery dataset to create.'
              }
            }
          },
          'legacy': False
        },
        'to': {
          'dataset': {
            'field': {
              'name': 'recipe_slug',
              'kind': 'string',
              'order': 4,
              'default': '',
              'description': 'Name of Google BigQuery dataset to create.'
            }
          },
          'view': 'Barnacle_User_Roles'
        }
      }
    },
    {
      'bigquery': {
        'auth': {
          'field': {
            'name': 'auth_write',
            'kind': 'authentication',
            'order': 1,
            'default': 'service',
            'description': 'Credentials used for writing data.'
          }
        },
        'from': {
          'query': "SELECT           R.*,           P.displayName AS partnerName,           A.displayName AS advertiserName,           FROM (           SELECT             queryId,             (SELECT CAST(value AS INT64) FROM UNNEST(R.params.filters) WHERE type = 'FILTER_PARTNER' LIMIT 1) AS partnerId,             (SELECT CAST(value AS INT64) FROM UNNEST(R.params.filters) WHERE type = 'FILTER_ADVERTISER' LIMIT 1) AS advertiserId,             R.schedule.frequency,             R.params.metrics,             R.params.type,             R.metadata.dataRange,             R.metadata.sendNotification,             DATE(TIMESTAMP_MILLIS(R.metadata.latestReportRunTimeMS)) AS latestReportRunTime,           FROM `{dataset}.DV_Reports` AS R) AS R           LEFT JOIN `{dataset}.DV_Partners` AS P           ON R.partnerId=P.partnerId           LEFT JOIN `{dataset}.DV_Advertisers` AS A           ON R.advertiserId=A.advertiserId         ",
          'parameters': {
            'dataset': {
              'field': {
                'name': 'recipe_slug',
                'kind': 'string',
                'order': 4,
                'default': '',
                'description': 'Name of Google BigQuery dataset to create.'
              }
            }
          },
          'legacy': False
        },
        'to': {
          'dataset': {
            'field': {
              'name': 'recipe_slug',
              'kind': 'string',
              'order': 4,
              'default': '',
              'description': 'Name of Google BigQuery dataset to create.'
            }
          },
          'view': 'Barnacle_Reports'
        }
      }
    }
  ]
}

DAG_FACTORY = DAG_Factory('barnacle_dv360', RECIPE, INPUTS)
DAG = DAG_FACTORY.generate()

if __name__ == "__main__":
  DAG_FACTORY.print_commandline()



###########################################################################
#
#  Copyright 2020 Google LLC
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      https://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
#
###########################################################################
"""Handler that executes { "drive":{...}} task in recipe JSON.

This script perfroms various drive operations like copy and create.
"""

from starthinker.util.project import project
from starthinker.util.drive import file_copy, file_create, file_delete


@project.from_parameters
def drive():
  if project.verbose:
    print('Drive')

  if 'delete' in project.task:
    if project.verbose:
      print('Drive Delete', project.task['delete'])
    file_delete(project.task['auth'], project.task['delete'])

  if 'copy' in project.task:
    if project.verbose:
      print('Drive Copy', project.task['copy']['source'],
            project.task['copy']['destination'])
    file_copy(project.task['auth'], project.task['copy']['source'],
              project.task['copy']['destination'])


if __name__ == '__main__':
  drive()



from starthinker.util.project import project
from starthinker.util.sheets import sheets_read


@project.from_parameters
def plx():
  # read peers from sheet
  rows = sheets_read(project.task['auth'], project.task['sheet']['url'],
                     project.task['sheet']['tab'],
                     project.task['sheet']['range'])

  wheres = []
  # construct where clause
  for row in rows:
    if not row:
      continue
    account, advertisers = row
    if account and advertisers:
      for advertiser in advertisers.split(','):
        wheres.append('(%s=%d AND %s=%d)' %
                      (project.task['field']['account'], int(account),
                       project.task['field']['advertiser'], int(advertiser)))
    elif advertisers:
      for advertiser in advertisers.split(','):
        wheres.append('%s=%d' %
                      (project.task['field']['advertiser'], int(advertiser)))
    elif account:
      wheres.append('%s=%d' % (project.task['field']['account'], int(account)))

  wheres = '\n  OR '.join(wheres)

  # print ( no API so STDOUT )
  print(project.task['query'].replace('[WHERE]', wheres).replace('; ', ';\n'))


if __name__ == '__main__':
  plx()



###########################################################################
#
#  Copyright 2020 Google LLC
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      https://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
#
###########################################################################
"""Command line to convert script with fields into a recipe with specific values.

Scripts are JSON templates that define a workflow.  Files in the format
script_*.json.
Each script can be converted to a recipe by replacing { field:{...}} elements
with
actual vaues.

This program reads {{ field:{...}} values from the script JSON, translates them
into
command line arguments, and uses those arguments to fill in the JSON {{
field:{...}} values
and produce a specific recipe as STDOUT.  Which can be piped to a file.

Example:

  `python starthinker/script/run.py scripts/script_dcm_to_bigquery.json`

  Will produce the following because it expects the arguments in the json
  script.

  ```
  usage: run.py [-h] json account report_id report_name dataset table
  run.py: error: too few arguments
  ```

  To see a detailed list of arguments run with the -h option:

  ```
  python starthinker/script/run.py scripts/script_dcm_to_bigquery.json -h
  usage: run.py [-h] json account report_id report_name dataset table

  positional arguments:
    json         JSON input recipe template file to onfigure.
    account      DCM network id.
    report_id    DCM report id from the UI.
    report_name  DCM report name, pass '' if using id instead.
    dataset      Dataset to be written to in BigQuery.
    table        Table to be written to in BigQuery.

  optional arguments:
    -h, --help   show this help message and exit
    --output JSON file to write resulting recipe.
  ```

  Then to turn the script into a recipe run:

  `python starthinker/script/run.py scripts/script_dcm_to_bigquery.json 7880
  1234567 "" "Test_Dataset" "Test_Table" > test_recipe.json`

  To perform the work of the script for the now filled in recipe:

  `python starthinker/all/run.py test_recipe.json`

"""

import sys
import json
import argparse

from starthinker.util.project import get_project
from starthinker.script.parse import json_get_fields, json_set_fields, json_set_instructions, json_set_description


def parser_add_field(parser, field):
  """Translates JOSN field specification into a command line argument.

    Args:
      parser: (ArgumentParser) An existing initalized argument parser to add
        fields to.
      field: (dict) A filed structured as: { "name":"???", "kind":"???",
        "default":???, "description":"???" }}

    Returns:
      Nothing.  Modifies parser in place.

    Raises:
      NotImplementedError: If field cannot be found.

  """

  if field['kind'] == 'string':
    parser.add_argument(
        '--' + field['name'],
        help=field.get('description', 'No instructions.'),
        default=field.get('default', ''))
  elif field['kind'] == 'email':
    parser.add_argument(
        '--' + field['name'],
        help=field.get('description', 'No instructions.'),
        default=field.get('default', ''))
  elif field['kind'] == 'authentication':
    parser.add_argument(
        '--' + field['name'],
        help=field.get('description', 'No instructions.'),
        default=field.get('default', ''),
        choices=['user', 'service'])
  elif field['kind'] == 'integer':
    parser.add_argument(
        '--' + field['name'],
        type=int,
        help=field.get('description', 'No instructions.'),
        default=field.get('default', ''))
  elif field['kind'] == 'boolean':
    parser.add_argument(
        '--' + field['name'],
        help=field.get('description', 'No instructions.'),
        action='store_%s' % str(field.get('default', 'false')).lower())
  elif field['kind'] == 'text':
    parser.add_argument(
        '--' + field['name'],
        help=field.get('description', 'No instructions.'),
        default=field.get('default', ''))
  elif field['kind'] == 'choice':
    parser.add_argument(
        '--' + field['name'],
        nargs='?',
        choices=field['choices'],
        help=field.get('description', 'No instructions'),
        default=field.get('default', ''))
  elif field['kind'] == 'timezone':
    parser.add_argument(
        '--' + field['name'],
        help=field.get('description', 'No instructions.'),
        default=field.get('default', ''))
  elif field['kind'] == 'json':
    raise NotImplementedError('JSON is not a suported field type')
  elif field['kind'] == 'integer_list':
    parser.add_argument(
        '--' + field['name'],
        type=int,
        nargs='+',
        help=field.get('description', 'No instructions.'),
        default=field.get('default', ''))
  elif field['kind'] == 'string_list':
    parser.add_argument(
        '--' + field['name'],
        nargs='+',
        help=field.get('description', 'No instructions.'),
        default=field.get('default', ''))
  else:
    raise NotImplementedError('%s is not a suported field type' % field)


def script_write(script, args, filepath=None):
  # insert fields into json
  json_set_fields(script, args)

  # insert variables into instructions
  json_set_instructions(script, args)

  # insert variables into description
  json_set_description(script, args)

  # produce output or run task
  if filepath:
    with open(filepath, 'w') as data_file:
      data_file.write(json.dumps(script, indent=2))
      print('JSON Written To: ', filepath)
  else:
    print(json.dumps(script, indent=2))


def script_interactive():
  args = {}

  from_json = sys.argv[1]
  to_json = sys.argv[2] if len(sys.argv) == 3 else ''

  script = get_project(sys.argv[1])

  # parse fields and constants into parameters
  fields = json_get_fields(script)

  print('-' * 60)
  if to_json:
    print('(1 of %d) From %s template create recipe: %s\n' %
          (len(fields), from_json, to_json))
  else:
    print('(1 of %d) Recipe file to create from %s template.\n' %
          (len(fields), sys.argv[1]))
    to_json = input('Full Path TO JSON File:')

  for count, field in enumerate(fields):
    print('')
    print('-' * 60)
    print('(%d of %d) %s%s%s' % (
        count + 2,
        len(fields),
        field['description'],
        ','.join(field['choices']) if 'choices' in field else '',
        (' Default to "%s" if blank.' %
         field['default']) if 'default' in field else '',
    ))
    args[field['name']] = input('%s ( %s ): ' % (field['name'], field['kind']))

    # remove blanks ( they should have defaults )
    if not args[field['name']]:
      del args[field['name']]

  script_write(script, args, to_json)


def script_commandline():

  script = get_project(sys.argv[1])

  # assemble parameters
  parser = argparse.ArgumentParser()
  parser.add_argument(
      'json', help='JSON recipe template to configure.', default=None)

  # parse fields and constants into parameters
  for field in json_get_fields(script):
    parser_add_field(parser, field)

  # run new arg parser with parameters from script
  args = vars(parser.parse_args())

  # always write to STDOUT, caller should rediect output to a JSON file
  script_write(script, args)


def main():
  # invalid command line
  if len(sys.argv) < 2:
    print('USAGE: run.py [json recipe template file path] [-h]')

  # only script, json and optional destination supplied, assume interactive
  elif len(sys.argv) == 2 or (len(sys.argv) == 3 and sys.argv[2] != '-h'):
    script_interactive()

  # parameters supplied, assume command line
  else:
    script_commandline()


if __name__ == '__main__':
  main()



###########################################################################
#
#  Copyright 2020 Google LLC
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      https://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
#
###########################################################################
"""Generic utilities that do not belong in any specific sub module.

Add general utility functions that are used across many modules.  Do
not add classes here.
"""


def flag_last(o):
  """Flags the last loop of an iterator.

  Consumes an iterator, buffers one instance so it can look ahead.
  Returns True on last iteration.

  Args:
    * o: An iterator instance.

  Returns:
    * A tuple of (True/False, iteration). Returns True, next on StopIteration.

  """

  it = o.__iter__()

  try:
    e = next(it)
  except StopIteration:
    return

  while True:
    try:
      nxt = next(it)
      yield (False, e)
      e = nxt
    except StopIteration:
      yield (True, e)
      break



###########################################################################
#
#  Copyright 2020 Google LLC
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      https://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
#
###########################################################################

LineItem_Schema = [
    {
        'name':
            'common_data',
        'type':
            'RECORD',
        'mode':
            'NULLABLE',
        'fields': [
            {
                'name': 'id',
                'type': 'INTEGER',
                'mode': 'NULLABLE',
            },
            {
                'name': 'name',
                'type': 'STRING',
                'mode': 'NULLABLE',
            },
            {
                'name': 'active',
                'type': 'BOOLEAN',
                'mode': 'NULLABLE',
            },
            {
                'name': 'integration_code',
                'type': 'STRING',
                'mode': 'NULLABLE',
            },
        ]
    },
    {
        'name': 'line_item_type',
        'type': 'INTEGER',
        'mode': 'NULLABLE',
    },
    {
        'name': 'insertion_order_id',
        'type': 'INTEGER',
        'mode': 'NULLABLE',
    },
    {
        'name': 'creative_ids',
        'type': 'INTEGER',
        'mode': 'REPEATED',
    },
    {
        'name': 'max_cpm_advertiser_micros',
        'type': 'INTEGER',
        'mode': 'NULLABLE',
    },
    {
        'name': 'performance_goal',
        'type': 'INTEGER',
        'mode': 'NULLABLE',
    },
    {
        'name': 'goal_advertiser_micros',
        'type': 'INTEGER',
        'mode': 'NULLABLE',
    },
    {
        'name':
            'partner_revenue_model',
        'type':
            'RECORD',
        'mode':
            'NULLABLE',
        'fields': [
            {
                'name': 'type',
                'type': 'INTEGER',
                'mode': 'NULLABLE',
            },
            {
                'name': 'amount_advertiser_micros',
                'type': 'INTEGER',
                'mode': 'NULLABLE',
            },
            {
                'name': 'media_cost_markup_percent_millis',
                'type': 'INTEGER',
                'mode': 'NULLABLE',
            },
            {
                'name': 'post_view_conversion_tracking_fraction',
                'type': 'FLOAT',
                'mode': 'NULLABLE',
            },
        ]
    },
    {
        'name':
            'cost_tracking_pixels',
        'type':
            'RECORD',
        'mode':
            'REPEATED',
        'fields': [
            {
                'name': 'pixel_id',
                'type': 'INTEGER',
                'mode': 'NULLABLE',
            },
            {
                'name': 'view_window_minutes',
                'type': 'INTEGER',
                'mode': 'NULLABLE',
            },
            {
                'name': 'click_window_minutes',
                'type': 'INTEGER',
                'mode': 'NULLABLE',
            },
        ]
    },
    {
        'name':
            'budget',
        'type':
            'RECORD',
        'mode':
            'NULLABLE',
        'fields': [
            {
                'name': 'start_time_usec',
                'type': 'INTEGER',
                'mode': 'NULLABLE',
            },
            {
                'name': 'end_time_usec',
                'type': 'INTEGER',
                'mode': 'NULLABLE',
            },
            {
                'name': 'max_impressions',
                'type': 'INTEGER',
                'mode': 'NULLABLE',
            },
            {
                'name': 'max_spend_advertiser_micros',
                'type': 'INTEGER',
                'mode': 'NULLABLE',
            },
            {
                'name': 'pacing_type',
                'type': 'INTEGER',
                'mode': 'NULLABLE',
            },
            {
                'name': 'pacing_max_impressions',
                'type': 'INTEGER',
                'mode': 'NULLABLE',
            },
            {
                'name': 'pacing_max_spend_advertiser_micros',
                'type': 'INTEGER',
                'mode': 'NULLABLE',
            },
            {
                'name': 'pacing_distribution',
                'type': 'INTEGER',
                'mode': 'NULLABLE',
            },
        ]
    },
    {
        'name':
            'frequency_cap',
        'type':
            'RECORD',
        'mode':
            'NULLABLE',
        'fields': [
            {
                'name': 'max_impressions',
                'type': 'INTEGER',
                'mode': 'NULLABLE',
            },
            {
                'name': 'time_unit',
                'type': 'INTEGER',
                'mode': 'NULLABLE',
            },
            {
                'name': 'time_range',
                'type': 'INTEGER',
                'mode': 'NULLABLE',
            },
        ]
    },
    {
        'name':
            'target_list',
        'type':
            'RECORD',
        'mode':
            'NULLABLE',
        'fields': [
            {
                'name':
                    'inventory_sources',
                'type':
                    'RECORD',
                'mode':
                    'REPEATED',
                'fields': [
                    {
                        'name': 'criteria_id',
                        'type': 'INTEGER',
                        'mode': 'NULLABLE',
                    },
                    {
                        'name': 'excluded',
                        'type': 'BOOLEAN',
                        'mode': 'NULLABLE',
                    },
                ]
            },
            {
                'name':
                    'geo_locations',
                'type':
                    'RECORD',
                'mode':
                    'REPEATED',
                'fields': [
                    {
                        'name': 'criteria_id',
                        'type': 'INTEGER',
                        'mode': 'NULLABLE',
                    },
                    {
                        'name': 'excluded',
                        'type': 'BOOLEAN',
                        'mode': 'NULLABLE',
                    },
                ]
            },
            {
                'name':
                    'ad_position',
                'type':
                    'RECORD',
                'mode':
                    'NULLABLE',
                'fields': [
                    {
                        'name': 'criteria_id',
                        'type': 'INTEGER',
                        'mode': 'NULLABLE',
                    },
                    {
                        'name': 'excluded',
                        'type': 'BOOLEAN',
                        'mode': 'NULLABLE',
                    },
                ]
            },
            {
                'name':
                    'net_speed',
                'type':
                    'RECORD',
                'mode':
                    'NULLABLE',
                'fields': [
                    {
                        'name': 'criteria_id',
                        'type': 'INTEGER',
                        'mode': 'NULLABLE',
                    },
                    {
                        'name': 'excluded',
                        'type': 'BOOLEAN',
                        'mode': 'NULLABLE',
                    },
                ]
            },
            {
                'name':
                    'browsers',
                'type':
                    'RECORD',
                'mode':
                    'NULLABLE',
                'fields': [
                    {
                        'name':
                            'union',
                        'type':
                            'RECORD',
                        'mode':
                            'REPEATED',
                        'fields': [
                            {
                                'name': 'criteria_id',
                                'type': 'INTEGER',
                                'mode': 'NULLABLE',
                            },
                            {
                                'name': 'parameter',
                                'type': 'STRING',
                                'mode': 'NULLABLE',
                            },
                        ]
                    },
                    {
                        'name': 'excluded',
                        'type': 'BOOLEAN',
                        'mode': 'NULLABLE',
                    },
                ]
            },
            {
                'name':
                    'device_criteria',
                'type':
                    'RECORD',
                'mode':
                    'REPEATED',
                'fields': [
                    {
                        'name': 'criteria_id',
                        'type': 'INTEGER',
                        'mode': 'NULLABLE',
                    },
                    {
                        'name': 'excluded',
                        'type': 'BOOLEAN',
                        'mode': 'NULLABLE',
                    },
                ]
            },
            {
                'name':
                    'languages',
                'type':
                    'RECORD',
                'mode':
                    'NULLABLE',
                'fields': [
                    {
                        'name':
                            'union',
                        'type':
                            'RECORD',
                        'mode':
                            'REPEATED',
                        'fields': [
                            {
                                'name': 'criteria_id',
                                'type': 'INTEGER',
                                'mode': 'NULLABLE',
                            },
                            {
                                'name': 'parameter',
                                'type': 'STRING',
                                'mode': 'NULLABLE',
                            },
                        ]
                    },
                    {
                        'name': 'excluded',
                        'type': 'BOOLEAN',
                        'mode': 'NULLABLE',
                    },
                ]
            },
            {
                'name':
                    'day_parting',
                'type':
                    'RECORD',
                'mode':
                    'NULLABLE',
                'fields': [
                    {
                        'name':
                            'union',
                        'type':
                            'RECORD',
                        'mode':
                            'REPEATED',
                        'fields': [
                            {
                                'name': 'criteria_id',
                                'type': 'INTEGER',
                                'mode': 'NULLABLE',
                            },
                            {
                                'name': 'parameter',
                                'type': 'STRING',
                                'mode': 'NULLABLE',
                            },
                        ]
                    },
                    {
                        'name': 'excluded',
                        'type': 'BOOLEAN',
                        'mode': 'NULLABLE',
                    },
                ]
            },
            {
                'name':
                    'audience_intersect',
                'type':
                    'RECORD',
                'mode':
                    'REPEATED',
                'fields': [
                    {
                        'name':
                            'union',
                        'type':
                            'RECORD',
                        'mode':
                            'REPEATED',
                        'fields': [
                            {
                                'name': 'criteria_id',
                                'type': 'INTEGER',
                                'mode': 'NULLABLE',
                            },
                            {
                                'name': 'parameter',
                                'type': 'STRING',
                                'mode': 'NULLABLE',
                            },
                        ]
                    },
                    {
                        'name': 'excluded',
                        'type': 'BOOLEAN',
                        'mode': 'NULLABLE',
                    },
                ]
            },
            {
                'name':
                    'keywords',
                'type':
                    'RECORD',
                'mode':
                    'REPEATED',
                'fields': [
                    {
                        'name': 'criteria_id',
                        'type': 'INTEGER',
                        'mode': 'NULLABLE',
                    },
                    {
                        'name': 'parameter',
                        'type': 'STRING',
                        'mode': 'NULLABLE',
                    },
                    {
                        'name': 'excluded',
                        'type': 'BOOLEAN',
                        'mode': 'NULLABLE',
                    },
                ]
            },
            {
                'name': 'kct_include_uncrawled_sites',
                'type': 'BOOLEAN',
                'mode': 'NULLABLE',
            },
            {
                'name':
                    'page_categories',
                'type':
                    'RECORD',
                'mode':
                    'REPEATED',
                'fields': [
                    {
                        'name': 'criteria_id',
                        'type': 'INTEGER',
                        'mode': 'NULLABLE',
                    },
                    {
                        'name': 'excluded',
                        'type': 'BOOLEAN',
                        'mode': 'NULLABLE',
                    },
                ]
            },
            {
                'name':
                    'universal_channels',
                'type':
                    'RECORD',
                'mode':
                    'REPEATED',
                'fields': [
                    {
                        'name': 'criteria_id',
                        'type': 'INTEGER',
                        'mode': 'NULLABLE',
                    },
                    {
                        'name': 'excluded',
                        'type': 'BOOLEAN',
                        'mode': 'NULLABLE',
                    },
                ]
            },
            {
                'name':
                    'sites',
                'type':
                    'RECORD',
                'mode':
                    'REPEATED',
                'fields': [
                    {
                        'name': 'criteria_id',
                        'type': 'INTEGER',
                        'mode': 'NULLABLE',
                    },
                    {
                        'name': 'parameter',
                        'type': 'STRING',
                        'mode': 'NULLABLE',
                    },
                    {
                        'name': 'excluded',
                        'type': 'BOOLEAN',
                        'mode': 'NULLABLE',
                    },
                ]
            },
            {
                'name':
                    'brand_safety',
                'type':
                    'RECORD',
                'mode':
                    'NULLABLE',
                'fields': [
                    {
                        'name': 'criteria_id',
                        'type': 'INTEGER',
                        'mode': 'NULLABLE',
                    },
                    {
                        'name': 'excluded',
                        'type': 'BOOLEAN',
                        'mode': 'NULLABLE',
                    },
                ]
            },
        ]
    },
    {
        'name':
            'partner_costs',
        'type':
            'RECORD',
        'mode':
            'NULLABLE',
        'fields': [
            {
                'name': 'cpm_fee_1_advertiser_micros',
                'type': 'INTEGER',
                'mode': 'NULLABLE',
            },
            {
                'name': 'cpm_fee_2_advertiser_micros',
                'type': 'INTEGER',
                'mode': 'NULLABLE',
            },
            {
                'name': 'cpm_fee_3_advertiser_micros',
                'type': 'INTEGER',
                'mode': 'NULLABLE',
            },
            {
                'name': 'cpm_fee_4_advertiser_micros',
                'type': 'INTEGER',
                'mode': 'NULLABLE',
            },
            {
                'name': 'cpm_fee_5_advertiser_micros',
                'type': 'INTEGER',
                'mode': 'NULLABLE',
            },
            {
                'name': 'media_fee_percent_1_millis',
                'type': 'INTEGER',
                'mode': 'NULLABLE',
            },
            {
                'name': 'media_fee_percent_2_millis',
                'type': 'INTEGER',
                'mode': 'NULLABLE',
            },
            {
                'name': 'media_fee_percent_3_millis',
                'type': 'INTEGER',
                'mode': 'NULLABLE',
            },
            {
                'name': 'media_fee_percent_4_millis',
                'type': 'INTEGER',
                'mode': 'NULLABLE',
            },
            {
                'name': 'media_fee_percent_5_millis',
                'type': 'INTEGER',
                'mode': 'NULLABLE',
            },
            {
                'name': 'cpm_fee_1_cost_type',
                'type': 'INTEGER',
                'mode': 'NULLABLE',
            },
            {
                'name': 'cpm_fee_2_cost_type',
                'type': 'INTEGER',
                'mode': 'NULLABLE',
            },
            {
                'name': 'cpm_fee_3_cost_type',
                'type': 'INTEGER',
                'mode': 'NULLABLE',
            },
            {
                'name': 'cpm_fee_4_cost_type',
                'type': 'INTEGER',
                'mode': 'NULLABLE',
            },
            {
                'name': 'cpm_fee_5_cost_type',
                'type': 'INTEGER',
                'mode': 'NULLABLE',
            },
            {
                'name': 'media_fee_percent_1_cost_type',
                'type': 'INTEGER',
                'mode': 'NULLABLE',
            },
            {
                'name': 'media_fee_percent_2_cost_type',
                'type': 'INTEGER',
                'mode': 'NULLABLE',
            },
            {
                'name': 'media_fee_percent_3_cost_type',
                'type': 'INTEGER',
                'mode': 'NULLABLE',
            },
            {
                'name': 'media_fee_percent_4_cost_type',
                'type': 'INTEGER',
                'mode': 'NULLABLE',
            },
            {
                'name': 'media_fee_percent_5_cost_type',
                'type': 'INTEGER',
                'mode': 'NULLABLE',
            },
            {
                'name': 'cpm_fee_1_bill_to_type',
                'type': 'INTEGER',
                'mode': 'NULLABLE',
            },
            {
                'name': 'cpm_fee_2_bill_to_type',
                'type': 'INTEGER',
                'mode': 'NULLABLE',
            },
            {
                'name': 'cpm_fee_3_bill_to_type',
                'type': 'INTEGER',
                'mode': 'NULLABLE',
            },
            {
                'name': 'cpm_fee_4_bill_to_type',
                'type': 'INTEGER',
                'mode': 'NULLABLE',
            },
            {
                'name': 'cpm_fee_5_bill_to_type',
                'type': 'INTEGER',
                'mode': 'NULLABLE',
            },
            {
                'name': 'media_fee_percent_1_bill_to_type',
                'type': 'INTEGER',
                'mode': 'NULLABLE',
            },
            {
                'name': 'media_fee_percent_2_bill_to_type',
                'type': 'INTEGER',
                'mode': 'NULLABLE',
            },
            {
                'name': 'media_fee_percent_3_bill_to_type',
                'type': 'INTEGER',
                'mode': 'NULLABLE',
            },
            {
                'name': 'media_fee_percent_4_bill_to_type',
                'type': 'INTEGER',
                'mode': 'NULLABLE',
            },
            {
                'name': 'media_fee_percent_5_bill_to_type',
                'type': 'INTEGER',
                'mode': 'NULLABLE',
            },
        ]
    },
]



# -*- coding: utf-8 -*-

###########################################################################
#
#  Copyright 2020 Google LLC
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      https://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
#
###########################################################################

from __future__ import unicode_literals

import os

from django.test import TestCase
from django.conf import settings

from starthinker_ui.account.tests import account_create
from starthinker_ui.project.models import Project


def project_create(share=''):

  with open(
      os.environ.get('STARTHINKER_SERVICE', 'MISSING RUN deploy.sh TO SET'),
      'r') as f:
    service = f.read()

  key = os.environ.get('STARTHINKER_API_KEY', 'MISSING RUN deploy.sh TO SET'),

  project = Project.objects.create(
      account=account_create(),
      service=service,
      key=key,
      share=share
  )

  return project


class ProjectTest(TestCase):

  def setUp(self):
    self.account = account_create()

    self.project_user = project_create()
    self.project_domain = project_create('domain')
    self.project_global = project_create('global')

  def test_ui_project(self):

    # not logged in ( blank )
    resp = self.client.get('/project/')
    self.assertEqual(resp.status_code, 200)

    self.client.force_login(
        self.account, backend=settings.AUTHENTICATION_BACKENDS[0])

    # logged in
    resp = self.client.get('/project/')
    self.assertEqual(resp.status_code, 200)

    self.assertContains(resp, self.project_user.identifier)
    self.assertContains(resp, self.project_domain.identifier)
    self.assertContains(resp, self.project_global.identifier)
    self.assertContains(resp, self.project_domain.share.upper())
    self.assertContains(resp, self.project_global.share.upper())

  def test_ui_project_edit(self):

    # not logged in ( blank )
    resp = self.client.get('/project/')
    self.assertEqual(resp.status_code, 200)

    self.client.force_login(self.account)

    # logged in
    resp = self.client.get('/project/edit/')
    self.assertEqual(resp.status_code, 200)

  def test_ui_recipe_edit(self):

    # not logged in ( redirect )
    resp = self.client.get('/recipe/edit/')
    self.assertEqual(resp.status_code, 302)

    self.client.force_login(
        self.account, backend=settings.AUTHENTICATION_BACKENDS[0])

    # logged in ( projects in form )
    resp = self.client.get('/recipe/edit/')
    self.assertEqual(resp.status_code, 200)
    self.assertContains(resp, self.project_user)
    self.assertContains(resp, self.project_domain)
    self.assertContains(resp, self.project_global)



###########################################################################
#
#  Copyright 2020 Google LLC
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      https://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
#
###########################################################################

from django.shortcuts import render
from django.template.loader import render_to_string

from starthinker_ui.recipe.scripts import Script


def help(request):
  # if open source then request will be null
  if request:
    return render(request, 'website/help.html', {})
  else:
    return render_to_string('website/help.html', {'external': True})


def solutions(request):
  # if open source then request will be null
  if request:
    return render(request, 'website/solutions.html', {})
  else:
    return render_to_string('website/solutions.html', {'external': True})


def solution(request, tag):
  script = Script(tag)
  # if open source then request will be null
  if request:
    return render(request, 'website/solution.html', {'script': script})
  else:
    return render_to_string('website/solution.html', {
        'script': script,
        'external': True
    })



###########################################################################
#
#  Copyright 2020 Google LLC
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      https://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
#
###########################################################################

import gzip

from starthinker.util.project import project
from starthinker.util.data import put_rows, get_rows
from starthinker.util.csv import excel_to_rows, csv_to_rows, rows_trim, rows_header_sanitize, column_header_sanitize
from starthinker.util.email import get_email_messages, get_email_links, get_email_attachments, get_subject, send_email
from starthinker.util.dbm import report_to_rows as dv360_report_to_rows, report_clean as dv360_report_clean
from starthinker.util.dcm import report_to_rows as cm_report_to_rows, report_clean as cm_report_clean, report_schema as cm_report_schema


def email_read():

  # process only most recent message
  try:
    message = next(
        get_email_messages(project.task['auth'], project.task['read']['from'],
                           project.task['read']['to'],
                           project.task['read'].get('subject', None)))
  except StopIteration as e:
    if project.verbose:
      print('NO EMAILS FOUND')
    exit()

  if project.verbose:
    print('EMAIL:', get_subject(message))

  files = []

  if project.task['read'].get('attachment'):
    files.extend(
        get_email_attachments(project.task['auth'], message,
                              project.task['read']['attachment']))

  if project.task['read'].get('link'):
    files.extend(
        get_email_links(
            project.task['auth'],
            message,
            project.task['read']['link'],
            download=True))

  for filename, data in files:

    if project.verbose:
      print('EMAIL FILENAME:', filename)

    # decompress if necessary
    if filename.endswith('.gz'):
      data = gzip.GzipFile(fileobj=data, mode='rb')
      filename = filename[:-3]

    # if excel file, save each sheet individually
    if filename.endswith('.xlsx'):

      for sheet, rows in excel_to_rows(data):
        rows = rows_trim(rows)
        rows = rows_header_sanitize(rows)
        put_rows(project.task['read']['out'].get('auth', project.task['auth']),
                 project.task['read']['out'], rows, sheet)

    # if CM report
    elif project.task['read']['from'] == 'noreply-cm@google.com':
      rows = cm_report_to_rows(data.read().decode())
      rows = cm_report_clean(rows)

      # if bigquery, remove header and determine schema
      schema = None
      if 'bigquery' in project.task['read']['out']:
        schema = cm_report_schema(next(rows))
        project.task['read']['out']['bigquery']['schema'] = schema
        project.task['read']['out']['bigquery']['skip_rows'] = 1

      put_rows(project.task['read']['out'].get('auth', project.task['auth']),
               project.task['read']['out'], rows)

    # if dv360 report
    elif project.task['read']['from'] == 'noreply-dv360@google.com':
      rows = dv360_report_to_rows(data.getvalue().decode())
      rows = dv360_report_clean(rows)
      put_rows(project.task['read']['out'].get('auth', project.task['auth']),
               project.task['read']['out'], rows)

    # if csv
    elif filename.endswith('.csv'):
      rows = csv_to_rows(data.read().decode())
      rows = rows_header_sanitize(rows)
      put_rows(project.task['read']['out'].get('auth', project.task['auth']),
               project.task['read']['out'], rows)

    else:
      if project.verbose:
        print('UNSUPPORTED FILE:', filename)


def email_send():
  if project.verbose:
    print('EMAIL SEND')

  send_email(
      'user',
      project.task['send']['from'],
      project.task['send']['to'],
      project.task['send'].get('cc', ''),
      project.task['send']['subject'],
      project.task['send']['text'],
      project.task['send']['html'],
      project.task['send']['attachment']['filename'],
      get_rows('user', project.task['send']['attachment']),
  )


@project.from_parameters
def email():
  if 'read' in project.task:
    email_read()
  elif 'send' in project.task:
    email_send()


if __name__ == '__main__':
  email()



###########################################################################
#
#  Copyright 2020 Google LLC
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      https://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
#
###########################################################################

'''
--------------------------------------------------------------

Before running this Airflow module...

  Install StarThinker in cloud composer ( recommended ):

    From Release: pip install starthinker
    From Open Source: pip install git+https://github.com/google/starthinker

  Or push local code to the cloud composer plugins directory ( if pushing local code changes ):

    source install/deploy.sh
    4) Composer Menu
    l) Install All

--------------------------------------------------------------

  If any recipe task has "auth" set to "user" add user credentials:

    1. Ensure an RECIPE['setup']['auth']['user'] = [User Credentials JSON]

  OR

    1. Visit Airflow UI > Admin > Connections.
    2. Add an Entry called "starthinker_user", fill in the following fields. Last step paste JSON from authentication.
      - Conn Type: Google Cloud Platform
      - Project: Get from https://github.com/google/starthinker/blob/master/tutorials/cloud_project.md
      - Keyfile JSON: Get from: https://github.com/google/starthinker/blob/master/tutorials/deploy_commandline.md#optional-setup-user-credentials

--------------------------------------------------------------

  If any recipe task has "auth" set to "service" add service credentials:

    1. Ensure an RECIPE['setup']['auth']['service'] = [Service Credentials JSON]

  OR

    1. Visit Airflow UI > Admin > Connections.
    2. Add an Entry called "starthinker_service", fill in the following fields. Last step paste JSON from authentication.
      - Conn Type: Google Cloud Platform
      - Project: Get from https://github.com/google/starthinker/blob/master/tutorials/cloud_project.md
      - Keyfile JSON: Get from: https://github.com/google/starthinker/blob/master/tutorials/cloud_service.md

--------------------------------------------------------------

Census Data Correlation

Correlate another table with US Census data.  Expands a data set dimensions by finding population segments that correlate with the master table.

  - Pre-requisite is Census Normalize, run that at least once.
  - Specify JOIN, PASS, SUM, and CORRELATE columns to build the correlation query.
  - Define the DATASET and TABLE for the joinable source. Can be a view.
  - Choose the significance level.  More significance usually means more NULL results, balance quantity and quality using this value.
  - Specify where to write the results.
  - <br>IMPORTANT:</b> If you use VIEWS, you will have to delete them manually if the recipe changes.

--------------------------------------------------------------

This StarThinker DAG can be extended with any additional tasks from the following sources:
  - https://google.github.io/starthinker/
  - https://github.com/google/starthinker/tree/master/dags

'''

from starthinker.airflow.factory import DAG_Factory

INPUTS = {
  'auth': 'service',  # Credentials used for writing data.
  'join': '',  # Name of column to join on, must match Census Geo_Id column.
  'pass': [],  # Comma seperated list of columns to pass through.
  'sum': [],  # Comma seperated list of columns to sum, optional.
  'correlate': [],  # Comma seperated list of percentage columns to correlate.
  'from_dataset': '',  # Existing BigQuery dataset.
  'from_table': '',  # Table to use as join data.
  'significance': '80',  # Select level of significance to test.
  'to_dataset': '',  # Existing BigQuery dataset.
  'type': 'table',  # Write Census_Percent as table or view.
}

RECIPE = {
  'tasks': [
    {
      'census': {
        'auth': {
          'field': {
            'name': 'auth',
            'kind': 'authentication',
            'order': 0,
            'default': 'service',
            'description': 'Credentials used for writing data.'
          }
        },
        'correlate': {
          'join': {
            'field': {
              'name': 'join',
              'kind': 'string',
              'order': 1,
              'default': '',
              'description': 'Name of column to join on, must match Census Geo_Id column.'
            }
          },
          'pass': {
            'field': {
              'name': 'pass',
              'kind': 'string_list',
              'order': 2,
              'default': [
              ],
              'description': 'Comma seperated list of columns to pass through.'
            }
          },
          'sum': {
            'field': {
              'name': 'sum',
              'kind': 'string_list',
              'order': 3,
              'default': [
              ],
              'description': 'Comma seperated list of columns to sum, optional.'
            }
          },
          'correlate': {
            'field': {
              'name': 'correlate',
              'kind': 'string_list',
              'order': 4,
              'default': [
              ],
              'description': 'Comma seperated list of percentage columns to correlate.'
            }
          },
          'dataset': {
            'field': {
              'name': 'from_dataset',
              'kind': 'string',
              'order': 5,
              'default': '',
              'description': 'Existing BigQuery dataset.'
            }
          },
          'table': {
            'field': {
              'name': 'from_table',
              'kind': 'string',
              'order': 6,
              'default': '',
              'description': 'Table to use as join data.'
            }
          },
          'significance': {
            'field': {
              'name': 'significance',
              'kind': 'choice',
              'order': 7,
              'default': '80',
              'description': 'Select level of significance to test.',
              'choices': [
                '80',
                '90',
                '98',
                '99',
                '99.5',
                '99.95'
              ]
            }
          }
        },
        'to': {
          'dataset': {
            'field': {
              'name': 'to_dataset',
              'kind': 'string',
              'order': 9,
              'default': '',
              'description': 'Existing BigQuery dataset.'
            }
          },
          'type': {
            'field': {
              'name': 'type',
              'kind': 'choice',
              'order': 10,
              'default': 'table',
              'description': 'Write Census_Percent as table or view.',
              'choices': [
                'table',
                'view'
              ]
            }
          }
        }
      }
    }
  ]
}

DAG_FACTORY = DAG_Factory('bigquery_census_correlate', RECIPE, INPUTS)
DAG = DAG_FACTORY.generate()

if __name__ == "__main__":
  DAG_FACTORY.print_commandline()



###########################################################################
#
#  Copyright 2020 Google LLC
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      https://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
#
###########################################################################
"""Handler that executes { "smartsheet":{...}} task in recipe JSON.

This script translates JSON instructions into operations on smartsheet
reporting.

"""

import re
import json

from smartsheet import Smartsheet

from starthinker.util.project import project
from starthinker.util.data import put_rows
from starthinker.util.csv import column_header_sanitize

SMARTSHEET_PAGESIZE = 10000
SMARTSHEET_DATE = re.compile(r'^\d{4}[-/_]\d{2}[-/_]\d{2}$')

SMARTSHEET_TYPES = {
    'AUTO_NUMBER': 'STRING',
    'CREATED_BY': 'STRING',
    'CREATED_DATE': 'DATE',
    'MODIFIED_BY': 'STRING',
    'MODIFIED_DATE': 'DATE',
    'ABSTRACT_DATETIME': 'TIMESTAMP',
    'CHECKBOX': 'STRING',
    'CONTACT_LIST': 'STRING',
    'DATE': 'DATE',
    'DATETIME': 'TIMESTAMP',
    'DURATION': 'STRING',
    'MULTI_CONTACT_LIST': 'STRING',
    'MULTI_PICKLIST': 'STRING',
    'PICKLIST': 'STRING',
    'PREDECESSOR': 'STRING',
    'TEXT_NUMBER': 'STRING'
}


def smartsheet_api(token):
  smart = Smartsheet(access_token=token)
  smart.errors_as_exceptions(True)
  return smart


def get_schema(sheet_or_report):
  return [{
      'name': column_header_sanitize(column.title),
      'type': SMARTSHEET_TYPES.get(str(column.type), 'STRING'),
      'mode': 'NULLABLE'
  } for column in sheet_or_report.columns]


def get_rows(sheet=None, report=None, header=True, link=True, key='id'):

  columns = {(column.id if sheet else column.virtual_id): {
      'title': column.title,
      'index': column.index,
      'type': str(column.type)
  } for column in (sheet or report).columns}

  if header:
    cells = [
        column['title']
        for column in sorted(columns.values(), key=lambda i: i['index'])
    ]
    if link:
      cells.insert(0, 'rowPermalink')
    yield cells

  for row in (sheet or report).rows:
    buffer = [None] * len(columns)
    for cell in row.cells:
      value = cell.value

      # correct the date column ( for reports it comes in as date and time even though column is DATE)
      if columns[cell.column_id if sheet else cell
                 .virtual_column_id]['type'] == 'DATE' and value is not None:
        if 'T' in value:
          value = value.split('T', 1)[0]
        if not SMARTSHEET_DATE.match(value):
          print('BAD DATE VALUE', value)
          value = None

      buffer[columns[cell.column_id if sheet else cell.virtual_column_id]
             ['index']] = value

    if link:
      buffer.insert(0, row.permalink)
    yield buffer


def get_sheet_schema(token, sheet, link=True):
  if project.verbose:
    print('SMARTSHEET SHEET SCHEMA', sheet)
  sheet_json = smartsheet_api(token).Sheets.get_sheet(
      sheet, include=('rowPermalink' if link else ''), page_size=0)
  return get_schema(sheet_json)


def get_report_schema(token, report, link=True):
  if project.verbose:
    print('SMARTSHEET SHEET SCHEMA', report)
  report_json = smartsheet_api(token).Reports.get_report(report, page_size=0)
  return get_schema(report_json)


def get_sheet_rows(token, sheet, link=True):
  if project.verbose:
    print('SMARTSHEET SHEET ROWS', sheet)
  sheet_json = smartsheet_api(token).Sheets.get_sheet(
      sheet, include=('rowPermalink' if link else ''))
  return get_rows(sheet=sheet_json, header=False, link=link)


def get_report_rows(token, report):
  if project.verbose:
    print('SMARTSHEET REPORT ROWS', report)

  count = 0
  total = 1
  page = 1

  api = smartsheet_api(token)
  while count < total:
    report_json = api.Reports.get_report(
        report, page_size=SMARTSHEET_PAGESIZE, page=page)
    total = report_json.total_row_count
    for row in get_rows(report=report_json, header=False, link=False):
      yield row
      count += 1
    page += 1


@project.from_parameters
def smartsheet():
  if project.verbose:
    print('SMARTSHEET')

  if 'sheet' in project.task:
    link = project.task.get('link', True)
    rows = get_sheet_rows(project.task['token'], project.task['sheet'], link)
    schema = get_sheet_schema(project.task['token'], project.task['sheet'],
                              link)

  elif 'report' in project.task:
    link = False
    rows = get_report_rows(project.task['token'], project.task['report'])
    schema = get_report_schema(project.task['token'], project.task['report'])

  else:
    raise NameError('Either report or sheet must be in the recipe json.')

  if 'bigquery' in project.task['out']:
    project.task['out']['bigquery'].setdefault('schema', schema)
    print('SCHEMA = %s' %
          json.dumps(project.task['out']['bigquery']['schema'], indent=2))

    if link and 'schema' in project.task['out']['bigquery']:
      project.task['out']['bigquery']['schema'].insert(0, {
          'name': 'rowPermalink',
          'type': 'STRING',
          'mode': 'NULLABLE'
      })

  if rows:
    put_rows(project.task['auth'], project.task['out'], rows)


if __name__ == '__main__':
  smartsheet()



###########################################################################
#
#  Copyright 2020 Google LLC
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      https://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
#
###########################################################################

'''
--------------------------------------------------------------

Before running this Airflow module...

  Install StarThinker in cloud composer ( recommended ):

    From Release: pip install starthinker
    From Open Source: pip install git+https://github.com/google/starthinker

  Or push local code to the cloud composer plugins directory ( if pushing local code changes ):

    source install/deploy.sh
    4) Composer Menu
    l) Install All

--------------------------------------------------------------

  If any recipe task has "auth" set to "user" add user credentials:

    1. Ensure an RECIPE['setup']['auth']['user'] = [User Credentials JSON]

  OR

    1. Visit Airflow UI > Admin > Connections.
    2. Add an Entry called "starthinker_user", fill in the following fields. Last step paste JSON from authentication.
      - Conn Type: Google Cloud Platform
      - Project: Get from https://github.com/google/starthinker/blob/master/tutorials/cloud_project.md
      - Keyfile JSON: Get from: https://github.com/google/starthinker/blob/master/tutorials/deploy_commandline.md#optional-setup-user-credentials

--------------------------------------------------------------

  If any recipe task has "auth" set to "service" add service credentials:

    1. Ensure an RECIPE['setup']['auth']['service'] = [Service Credentials JSON]

  OR

    1. Visit Airflow UI > Admin > Connections.
    2. Add an Entry called "starthinker_service", fill in the following fields. Last step paste JSON from authentication.
      - Conn Type: Google Cloud Platform
      - Project: Get from https://github.com/google/starthinker/blob/master/tutorials/cloud_project.md
      - Keyfile JSON: Get from: https://github.com/google/starthinker/blob/master/tutorials/cloud_service.md

--------------------------------------------------------------

Monthly Budget Mover

Apply the previous month's budget/spend delta to the current month.  Aggregate up the budget and spend from the previous month of each category declared then apply the delta of the spend and budget equally to each Line Item under that Category.

  - No changes made can be made in DV360 from the start to the end of this process
  - Make sure there is budget information for the current and previous month's IOs in DV360
  - Make sure the provided spend report has spend data for every IO in the previous month
  - Spend report must contain 'Revenue (Adv Currency)' and 'Insertion Order ID'
  - There are no duplicate IO Ids in the categories outlined below
  - This process must be ran during the month of the budget it is updating
  - If you receive a 502 error then you must separate your jobs into two, because there is too much information being pulled in the sdf
  - Manually run this job
  - Once the job has completed go to the table for the new sdf and export to a csv
  - Take the new sdf and upload it into DV360

--------------------------------------------------------------

This StarThinker DAG can be extended with any additional tasks from the following sources:
  - https://google.github.io/starthinker/
  - https://github.com/google/starthinker/tree/master/dags

'''

from starthinker.airflow.factory import DAG_Factory

INPUTS = {
  'recipe_timezone': 'America/Los_Angeles',  # Timezone for report dates.
  'recipe_name': '',  # 
  'auth_write': 'service',  # Credentials used for writing data.
  'auth_read': 'user',  # Credentials used for reading data.
  'partner_id': '',  # The sdf file types.
  'budget_categories': '{}',  # A dictionary to show which IO Ids go under which Category. {"CATEGORY1":[12345,12345,12345], "CATEGORY2":[12345,12345]}
  'filter_ids': [],  # Comma separated list of filter ids for the request.
  'excluded_ios': '',  # A comma separated list of Inserion Order Ids that should be exluded from the budget calculations
  'version': '5',  # The sdf version to be returned.
  'is_colab': True,  # Are you running this in Colab? (This will store the files in Colab instead of Bigquery)
  'dataset': '',  # Dataset that you would like your output tables to be produced in.
}

RECIPE = {
  'setup': {
    'day': [
      'Mon',
      'Tue',
      'Wed',
      'Thu',
      'Fri',
      'Sat',
      'Sun'
    ],
    'hour': [
      2,
      4,
      6,
      8,
      10,
      12,
      14,
      16,
      18,
      20,
      22,
      24
    ]
  },
  'tasks': [
    {
      'dataset': {
        'description': 'Create a dataset where data will be combined and transfored for upload.',
        'auth': {
          'field': {
            'name': 'auth_write',
            'kind': 'authentication',
            'order': 1,
            'default': 'service',
            'description': 'Credentials used for writing data.'
          }
        },
        'dataset': {
          'field': {
            'name': 'dataset',
            'kind': 'string',
            'order': 1,
            'description': 'Place where tables will be created in BigQuery.'
          }
        }
      }
    },
    {
      'dbm': {
        'auth': {
          'field': {
            'name': 'auth_read',
            'kind': 'authentication',
            'order': 1,
            'default': 'user',
            'description': 'Credentials used for reading data.'
          }
        },
        'report': {
          'timeout': 90,
          'filters': {
            'FILTER_ADVERTISER': {
              'values': {
                'field': {
                  'name': 'filter_ids',
                  'kind': 'integer_list',
                  'order': 7,
                  'default': '',
                  'description': 'The comma separated list of Advertiser Ids.'
                }
              }
            }
          },
          'body': {
            'timezoneCode': {
              'field': {
                'name': 'recipe_timezone',
                'kind': 'timezone',
                'description': 'Timezone for report dates.',
                'default': 'America/Los_Angeles'
              }
            },
            'metadata': {
              'title': {
                'field': {
                  'name': 'recipe_name',
                  'kind': 'string',
                  'prefix': 'Monthly_Budget_Mover_',
                  'order': 1,
                  'description': 'Name of report in DV360, should be unique.'
                }
              },
              'dataRange': 'PREVIOUS_MONTH',
              'format': 'CSV'
            },
            'params': {
              'type': 'TYPE_GENERAL',
              'groupBys': [
                'FILTER_ADVERTISER_CURRENCY',
                'FILTER_INSERTION_ORDER'
              ],
              'metrics': [
                'METRIC_REVENUE_ADVERTISER'
              ]
            }
          }
        },
        'delete': False
      }
    },
    {
      'monthly_budget_mover': {
        'auth': 'user',
        'is_colab': {
          'field': {
            'name': 'is_colab',
            'kind': 'boolean',
            'default': True,
            'order': 7,
            'description': 'Are you running this in Colab? (This will store the files in Colab instead of Bigquery)'
          }
        },
        'report_name': {
          'field': {
            'name': 'recipe_name',
            'kind': 'string',
            'prefix': 'Monthly_Budget_Mover_',
            'order': 1,
            'description': 'Name of report in DV360, should be unique.'
          }
        },
        'budget_categories': {
          'field': {
            'name': 'budget_categories',
            'kind': 'json',
            'order': 3,
            'default': '{}',
            'description': 'A dictionary to show which IO Ids go under which Category. {"CATEGORY1":[12345,12345,12345], "CATEGORY2":[12345,12345]}'
          }
        },
        'excluded_ios': {
          'field': {
            'name': 'excluded_ios',
            'kind': 'integer_list',
            'order': 4,
            'description': 'A comma separated list of Inserion Order Ids that should be exluded from the budget calculations'
          }
        },
        'sdf': {
          'auth': 'user',
          'version': {
            'field': {
              'name': 'version',
              'kind': 'choice',
              'order': 6,
              'default': '5',
              'description': 'The sdf version to be returned.',
              'choices': [
                'SDF_VERSION_5',
                'SDF_VERSION_5_1'
              ]
            }
          },
          'partner_id': {
            'field': {
              'name': 'partner_id',
              'kind': 'integer',
              'order': 1,
              'description': 'The sdf file types.'
            }
          },
          'file_types': 'INSERTION_ORDER',
          'filter_type': 'FILTER_TYPE_ADVERTISER_ID',
          'read': {
            'filter_ids': {
              'single_cell': True,
              'values': {
                'field': {
                  'name': 'filter_ids',
                  'kind': 'integer_list',
                  'order': 4,
                  'default': [
                  ],
                  'description': 'Comma separated list of filter ids for the request.'
                }
              }
            }
          },
          'time_partitioned_table': False,
          'create_single_day_table': False,
          'dataset': {
            'field': {
              'name': 'dataset',
              'kind': 'string',
              'order': 6,
              'default': '',
              'description': 'Dataset to be written to in BigQuery.'
            }
          },
          'table_suffix': ''
        },
        'out_old_sdf': {
          'bigquery': {
            'dataset': {
              'field': {
                'name': 'dataset',
                'kind': 'string',
                'order': 8,
                'default': '',
                'description': 'Dataset that you would like your output tables to be produced in.'
              }
            },
            'table': {
              'field': {
                'name': 'recipe_name',
                'kind': 'string',
                'prefix': 'SDF_OLD_',
                'description': ''
              }
            },
            'schema': [
            ],
            'skip_rows': 0,
            'disposition': 'WRITE_TRUNCATE'
          },
          'file': '/content/old_sdf.csv'
        },
        'out_new_sdf': {
          'bigquery': {
            'dataset': {
              'field': {
                'name': 'dataset',
                'kind': 'string',
                'order': 8,
                'default': '',
                'description': 'Dataset that you would like your output tables to be produced in.'
              }
            },
            'table': {
              'field': {
                'name': 'recipe_name',
                'kind': 'string',
                'prefix': 'SDF_NEW_',
                'description': ''
              }
            },
            'schema': [
            ],
            'skip_rows': 0,
            'disposition': 'WRITE_TRUNCATE'
          },
          'file': '/content/new_sdf.csv'
        },
        'out_changes': {
          'bigquery': {
            'dataset': {
              'field': {
                'name': 'dataset',
                'kind': 'string',
                'order': 8,
                'default': '',
                'description': 'Dataset that you would like your output tables to be produced in.'
              }
            },
            'table': {
              'field': {
                'name': 'recipe_name',
                'kind': 'string',
                'prefix': 'SDF_BUDGET_MOVER_LOG_',
                'description': ''
              }
            },
            'schema': [
            ],
            'skip_rows': 0,
            'disposition': 'WRITE_TRUNCATE'
          },
          'file': '/content/log.csv'
        }
      }
    }
  ]
}

DAG_FACTORY = DAG_Factory('monthly_budget_mover', RECIPE, INPUTS)
DAG = DAG_FACTORY.generate()

if __name__ == "__main__":
  DAG_FACTORY.print_commandline()



###########################################################################
#
#  Copyright 2020 Google LLC
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      https://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
#
###########################################################################
"""The core singleton class of StarThinker that translates json to python.

Project loads JSON and parameters and combines them for execturion.  It handles
three important concepts:

  1. Load the JSON and make all task parameters available to python scripts.
  2. Load authentication, all three parameters are optional if scripts do not
     use them.  The following parameters can be passed for authentication.

    user.json - user credentials json ( generated from client ), is refreshed
                by StarThinker as required.  Can be provided as a local path
                or a Cloud Bucket Storage path for distributed jobs.

    service.json - service credentials json ( generated from cloud project ).
                   Passed as a local path or an embedded json object for
                   distributed jobs.

    client.json - client credentials json ( generated from cloud project ).
                  Also require a user json path which will be written to after
                  client authnetication.  Once authenticated this client is not
                  required.

    Credentials can be specified in one of three ways for maximum flexibility:

    A. Specify credentials on command line (highest priority if given)
       --user / -u = user credentials path
       --client / -c = client credentials path (requires user credentials path)
       --service / -s = service credentials path

    B. Define credentials paths in JSON (overruled by command line)
       In each json file create the following entry (client, or user, or
       service)

         ```
         {
           "setup":{
             "id":"[cloud project id]",
             "auth":{
               "client":"[/home/.credentials/hello-world_client.json]",
               "service":"[/home/.credentials/hello-world_service.json]",
               "user":"[/home/.credentials/hello-world_user.json]"
             }
          },
         }
         ```

    C. Use default credentials ( lowest priority, last resort )
       If neither the json not the command line provides a path, the
       environmental variable GOOGLE_APPLICATION_CREDENTIALS will be
       used for service accounts.  It is created by google cloud
       utilities.
"""

import os
import re
import json
import pytz
import argparse
import textwrap

from datetime import datetime
from importlib import import_module

from starthinker.util.debug import starthinker_trace_start


EXIT_ERROR = 1
EXIT_SUCCESS = 0
RE_UUID = re.compile(r'(\s*)("setup"\s*:\s*{)')


def get_project(filepath):
  """Loads json for Project Class.  Intended for this module only.

  Available as helper.

     Able to load JSON with newlines ( strips all newlines before load ).

     Args:
      - filepath: (string) The local file path to the recipe json file to load.

     Returns:
      Json of recipe file.
  """

  with open(filepath) as data_file:
    data = data_file.read().replace('\n', ' ')

    try:
      return json.loads(data)

    except ValueError as e:
      with open(filepath) as data_file:
        pos = 0
        for count, line in enumerate(data_file.readlines(), 1):
          pos += len(
              line
          )  # do not add newlines, e.pos was run on a version wher enewlines were removed
          if pos >= e.pos:
            e.lineno = count
            e.pos = pos
            e.args = (
                'JSON ERROR: %s LINE: %s CHARACTER: %s ERROR: %s LINE: %s' %
                (filepath, count, pos - e.pos, str(e.msg), line.strip()),)
            raise


def utc_to_timezone(timestamp, timezone):
  if timestamp:
    return timestamp.replace(tzinfo=pytz.utc).astimezone(
        pytz.timezone(timezone))
  else:
    return None


def is_scheduled(recipe, task={}):
  """Wrapper for day_hour_scheduled that returns True if current time zone safe hour is in recipe schedule.

     Used as a helper for any cron job running projects.  Keeping this logic in
     project
     helps avoid time zone detection issues and scheduling discrepencies between
     machines.

    Args:
      - recipe: (Recipe JSON) The JSON of a recipe.
      - task: ( dictionary / JSON ) The specific task being considered for
        execution.

    Returns:
      - True if task is scheduled to run current hour, else False.
  """

  tz = pytz.timezone(
    recipe.get('setup', {}).get('timezone', 'America/Los_Angeles')
  )
  now_tz = datetime.now(tz)
  day_tz = now_tz.strftime('%a')
  hour_tz = now_tz.hour

  days = recipe.get('setup', {}).get('day', [])
  hours = [
    int(h) for h in task.get(
      'hour', recipe.get('setup', {}).get('hour', [])
    )
  ]

  if days == [] or day_tz in days:
    if hours == [] or hour_tz in hours:
      return True

  return False


class project:
  """A singleton that represents the loaded recipe within python scripts.

  All access to json scripts within StarThinker must pass through the project
  class.  It handles parameters, time zones, permissions management, and
  scheduling overhead. Task function name must match JSON task name.

  Project is meant as the entry point into all StarThinker scripts as follows:

    ```
    from starthinker.util.project import project

    @project.from_parameters
    def task():
      pass # execute work using project.* as data from json

    if __name__ == "__main__":
      task()
    ```

  Project is meant to be used by a helper.

    ```
    import argparse
    from starthinker.util.project import project

    if __name__ == "__main__":

      # custom parameters
      parser = argparse.ArgumentParser()
      parser.add_argument('custom', help='custom parameter to be added to
      standard project set.')

      # initialize project
      project.from_commandline(parser=parser)

      # access arguments
      auth = 'service' if project.args.service else 'user'
      print(project.args.custom)
    ```

  Project can also be initialized directly for non-json tasks:

    ```
    from starthinker.util.project import project

    if __name__ == "__main__":
      var_recipe = '/somepath/recipe.json'
      var_user = '/somepath/user.json'
      var_service = '/somepath/service.json'

      project.initialize(_recipe=var_recipe, _user=var_user,
      _service=var_service, _verbose=True)
    ```

  Attributes:  Dealing with authentication...
      - project: (string) The Google Cloud project id.
      - user: (string) Path to the user credentials json file.  It can also be a
      Google Cloud Bucket path when passed to the class directly.
      - service: (string) Path to the service credentials json file.  It can
      also be a json object when passed to the project class directly.
      - client: (string) Path to the client credentials json file.  It can only
      be a local path.  Dealing with execution data...
      - instance: (integer) When executing all tasks, it is the one based index
      offset of the task to run.
      - date: (date) A specific date or 'TODAY', which is changed to today's
      date, passed to python scripts for reference.
      - hour: (integer) When executing all tasks, it is the hour if spefified
      for a task to execute.  Dealing with debugging...
      - verbose: (boolean) Prints all debug information in StarThinker code.
      See: if project.verbose: print '...'.
      - force: (boolean) For recipes with specific task hours, forces all tasks
      to run regardless of hour specified.
  """

  args = None
  recipe = None
  task = None
  instance = None
  filepath = None
  verbose = False
  force = False
  function = None
  timezone = None
  now = None
  date = None
  hour = None

  @classmethod
  def from_commandline(cls, _task=None, parser=None, arguments=None):
    """Used in StarThinker scripts as entry point for command line calls.

    Loads json for execution.

    Usage example:

    ```
    import argparse
    from starthinker.util.project import project

    if __name__ == "__main__":

      # custom parameters
      parser = argparse.ArgumentParser()
      parser.add_argument('custom', help='custom parameter to be added to
      standard project set.')

      # initialize project
      project.from_commandline(parser=parser)

      # access arguments
      auth = 'service' if project.args.service else 'user'
      print(project.args.custom)
    ```

    Args:
      - parser: (ArgumentParser) optional custom argument parser ( json argument
        becomes optional if not None )
      - arguments: (String) optional list of parameters to use when invoking
        project ( defaults to ALL if set to None )

    Returns:
      Nothing, this manipulates a singleton object.  All calls to project.*
      result in the same object.

    """

    if parser is None:
      parser = argparse.ArgumentParser(
          formatter_class=argparse.RawDescriptionHelpFormatter,
          description=textwrap.dedent("""\
          Command line to execute all tasks in a recipe once. ( Common Entry Point )

          This script dispatches all the tasks in a JSON recipe to handlers in sequence.
          For each task, it calls a subprocess to execute the JSON instructions, waits
          for the process to complete and dispatches the next task, until all tasks are
          complete or a critical failure ( exception ) is raised.

          If an exception is raised in any task, all following tasks are not executed by design.

          Example: python run.py [path to recipe file] --force
          Caution: This script does NOT check if the last job finished, potentially causing overruns.
          Notes:
            - To avoid running the entire script when debugging a single task, the command line
              can easily replace "all" with the name of any "task" in the json.  For example
              python all/run.py project/sample/say_hello.json

            - Can be easily replaced with the following to run only the "hello" task:
              python task/hello/run.py project/sample/say_hello.json

            - Or specified further to run only the second hello task:
              python task/hello/run.py project/sample/say_hello.json -i 2

      """))
      if arguments is None or '-j' in arguments:
        parser.add_argument('json', help='Path to recipe json file to load.')
    else:
      if arguments is None or '-j' in arguments:
        parser.add_argument(
            '--json', '-j', help='Path to recipe json file to load.')

    if arguments is None or '-p' in arguments:
      parser.add_argument(
          '--project',
          '-p',
          help='Cloud ID of Google Cloud Project.',
          default=None)
    if arguments is None or '-k' in arguments:
      parser.add_argument(
          '--key',
          '-k',
          help='API Key of Google Cloud Project.',
          default=None)
    if arguments is None or '-u' in arguments:
      parser.add_argument(
          '--user',
          '-u',
          help='Path to USER credentials json file.',
          default=None)
    if arguments is None or '-s' in arguments:
      parser.add_argument(
          '--service',
          '-s',
          help='Path to SERVICE credentials json file.',
          default=None)
    if arguments is None or '-c' in arguments:
      parser.add_argument(
          '--client',
          '-c',
          help='Path to CLIENT credentials json file.',
          default=None)

    if arguments is None or '-i' in arguments:
      parser.add_argument(
          '--instance',
          '-i',
          help='Instance number of the task to run ( for tasks with same name ) starting at 1.',
          default=1,
          type=int)

    if arguments is None or '-v' in arguments:
      parser.add_argument(
          '--verbose',
          '-v',
          help='Print all the steps as they happen.',
          action='store_true')
    if arguments is None or '-f' in arguments:
      parser.add_argument(
          '--force',
          '-force',
          help='Not used but included for compatiblity with another script.',
          action='store_true')

    if arguments is None or '-tp' in arguments:
      parser.add_argument(
          '--trace_print',
          '-tp',
          help='Execution trace written to stdout.',
          action='store_true')
    if arguments is None or '-tf' in arguments:
      parser.add_argument(
          '--trace_file',
          '-tf',
          help='Execution trace written to file.',
          action='store_true')

    cls.args = parser.parse_args()

    # initialize the project singleton with passed in parameters
    cls.initialize(
        get_project(cls.args.json) if hasattr(cls.args, 'json') else {},
        _task,
        getattr(cls.args, 'instance', None),
        getattr(cls.args, 'project', None),
        getattr(cls.args, 'user', None),
        getattr(cls.args, 'service', None),
        getattr(cls.args, 'client', None),
        getattr(cls.args, 'json', None),
        getattr(cls.args, 'key', None),
        getattr(cls.args, 'verbose', None),
        getattr(cls.args, 'force', None),
        getattr(cls.args, 'trace_print', None),
        getattr(cls.args, 'trace_file', None)
    )

  @classmethod
  def get_task_index(cls):
    i = 0
    for c, t in enumerate(cls.recipe.get('tasks', [])):
      if next(iter(t.keys())) == cls.function:
        i += 1
        if i == cls.instance:
          return c
    return None


  @classmethod
  def get_task(cls):
    #if cls.task is None:
    i = cls.get_task_index()
    cls.task = None if i is None else next(
        iter(cls.recipe['tasks'][i].values()))
    return cls.task


  @classmethod
  def get_tasks(cls, recipe=None):
    instances = {}
    for task in (recipe or cls.recipe['tasks']):
      script, task = next(iter(task.items()))

      # count instance per task
      instances.setdefault(script, 0)
      instances[script] += 1

      if _force or cls.force or is_scheduled(cls.recipe, task):
        yield task, instances[script]
      else:
        print(
          'Schedule Skipping: add --force to ignore schedule or run specific task handler'
        )


  @classmethod
  def set_task(cls, function, parameters):
    if cls.task is None:
      cls.recipe['tasks'].append({function: parameters})
      cls.function = function
      cls.task = parameters
      cls.instance = 1
    else:
      i = cls.get_task_index()
      cls.recipe['tasks'][i] = {function: parameters}
      cls.function = function
      cls.task = parameters
      cls.instance = sum([
          1 for c, t in enumerate(cls.recipe['tasks'])
          if t == function and c <= i
      ])

  @staticmethod
  def from_parameters(func):
    """Initializes a project singleton for execution by a task.

    Either loads parameters (recipe, instance) passed to task programatically,
    or if no parameters passed attmepts to load them from the command line.
    Uses decorator pattern, task name is inferred from function ebing decorated.

    Args:
      - recipe: (dict) JSON object representing the project ( setup plus at
        least one task )
      - instance: (integer) numeric offset of task to run if multiple calls to
        thsi task exist
    """

    def from_parameters_wrapper(recipe=None, instance=1):
      if recipe:
        project.initialize(
            _recipe=recipe, _task=func.__name__, _instance=instance)
      else:
        project.from_commandline(func.__name__)
      func()

    return from_parameters_wrapper

  @classmethod
  def initialize(cls,
                 _recipe={},
                 _task=None,
                 _instance=1,
                 _project=None,
                 _user=None,
                 _service=None,
                 _client=None,
                 _filepath=None,
                 _key=None,
                 _verbose=False,
                 _force=False,
                 _trace_print=False,
                 _trace_file=False):
    """Used in StarThinker scripts as programmatic entry point.

    Set up the project singleton for execution of a task, be sure to mimic
    defaults in helper
    this function loads credentials from various source ( command line argument,
    json, default credentials )
    it also sets up time zone aware date and various helper flags such as force
    and verbose.

    Usage example:
    ```
       from starthinker.util.project import project

       if __name__ == "__main__":
         user = 'user.json'
         service = 'service.json'
         recipe = {'setup':..., 'tasks':.... }
         project.initialize(
           _recipe=recipe,
           _user=user,
           _service=service,
           _verbose=True
         )
    ```

    Args:
      - _recipe: (dict) JSON object representing the project ( setup plus at
        least one task )
      - _task: (string) Task name form recipe json task list to execute.
      - _instance: (integer) See module description.
      - _project: (string) See module description.
      - _user: (string) See module description.
      - _service: (string) See module description.
      - _client: (string) See module description.
      - _key: (string) See module description.
      - _verbose: (boolean) See module description.
      - _force: (boolean) See module description.
      - _trace_print: (boolean) True if writing execution trace to stdout.
      - _trace_file: (boolean) True if writing execution trace to file ( see
        config.py ).

    Returns:
      Nothing, this manipulates a singleton object.  All calls to project.*
      result in the same object.
    """

    starthinker_trace_start(_trace_print, _trace_file)

    cls.recipe = _recipe
    cls.function = _task
    cls.instance = _instance
    cls.force = _force

    # populates the task variable based on function and instance
    cls.get_task()

    cls.verbose = _verbose
    cls.filepath = _filepath

    # add setup to json if not provided and loads command line credentials if given
    if 'setup' not in cls.recipe:
      cls.recipe['setup'] = {}
    if 'auth' not in cls.recipe['setup']:
      cls.recipe['setup']['auth'] = {}
    if _service:
      cls.recipe['setup']['auth']['service'] = _service
    if _client:
      cls.recipe['setup']['auth']['client'] = _client
    # if user explicity specified by command line
    if _user:
      cls.recipe['setup']['auth']['user'] = _user
    # or if user not given, then try default credentials ( disabled security risk to assume on behalf of recipe )
    #elif not cls.recipe['setup']['auth'].get('user'):
    #  cls.recipe['setup']['auth']['user'] = os.environ.get('GOOGLE_APPLICATION_CREDENTIALS', None)

    # if project id given, use it
    if _project:
      cls.recipe['setup']['id'] = _project

    if _key:
      cls.recipe['setup']['key'] = _key

    # TBD: if project id not given, use service credentials
    #elif not cls.recipe['setup'].get('id') and cls.recipe['setup']['auth'].get('service'):
    # TBD: if project id not given, use client credentials
    #elif not cls.recipe['setup'].get('id') and cls.recipe['setup']['auth'].get('client'):

    cls.id = cls.recipe['setup'].get('id')
    cls.key = cls.recipe['setup'].get('key')

    # find date based on timezone
    cls.timezone = pytz.timezone(
      cls.recipe['setup'].get(
        'timezone',
        'America/Los_Angeles'
      )
    )
    cls.now = datetime.now(cls.timezone)
    cls.date = cls.now.date()
    cls.hour = cls.now.hour

    if cls.verbose:
      print('TASK:', _task or 'all')
      print('DATE:', cls.now.date())
      print('HOUR:', cls.now.hour)

  @classmethod
  def execute(cls, _force=False):
    """Run all the tasks in a project in one sequence.

    ```
    from starthinker.util.project import project

    if __name__ == "__main__":
      var_user = '[USER CREDENTIALS JSON STRING OR PATH]'
      var_service = '[SERVICE CREDENTIALS JSON STRING OR PATH]'
      var_recipe = {
        "tasks":[
          { "dataset":{
            "auth":"service",
            "dataset":"Test_Dataset"
          }}
        ]
      }

      project.initialize(
        _recipe=var_recipe,
        _user=var_user,
        _service=var_service,
        _verbose=True
      )
      project.execute()
    ```

    For a full list of tasks see: scripts/*.json

    """

    returncode = EXIT_SUCCESS
    instances = {}
    for task in cls.recipe['tasks']:
      script, task = next(iter(task.items()))

      # count instance per task
      instances.setdefault(script, 0)
      instances[script] += 1

      print('RUNNING TASK:', '%s %d' % (script, instances[script]))

      if _force or cls.force or is_scheduled(cls.recipe, task):
        try:
          python_callable = getattr(
            import_module('starthinker.task.%s.run' % script),
            script
          )
          python_callable(cls.recipe, instances[script])
        except Exception as e:
          print(str(e))
          returncode = EXIT_ERROR
      else:
        print(
          'Schedule Skipping: add --force to ignore schedule or run specific task handler'
        )

    return returncode



###########################################################################
#
#  Copyright 2020 Google LLC
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      https://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
#
###########################################################################

import json
from time import sleep

from django.core.management.base import BaseCommand, CommandError

from starthinker.util.project import project
from starthinker.util.email import send_email
from starthinker.util.email.template import EmailTemplate
from starthinker_ui.account.models import Account
"""python starthinker_ui/manage.py newsletter --from kenjora@google.com --template starthinker/task/newsletter/internal/2019_12_03.json --to kenjora@google.com"""


class Command(BaseCommand):
  help = 'Send email status to users'

  def add_arguments(self, parser):
    parser.add_argument(
        '--template',
        action='store',
        dest='template',
        help='Template to use for email construction.',
    )

    parser.add_argument(
        '--from',
        action='store',
        dest='email_from',
        help='Email to send from ( must match an account).',
    )

    parser.add_argument(
        '--to',
        action='store',
        dest='email_to',
        default=None,
        help='Email to send to ( if not givem sends to all ).',
    )

    parser.add_argument(
        '--ignore',
        action='store',
        nargs='+',
        dest='ignore',
        default=[],
        help='Email to send to ( if not givem sends to all ).',
    )

    parser.add_argument(
        '--test',
        action='store_true',
        dest='test',
        default=False,
        help='Test print emails instead of sending them.',
    )

  def handle(self, *args, **kwargs):

    user = None
    accounts = Account.objects.all()

    # find user to send from
    for account in accounts:
      if account.email == kwargs['email_from']:
        user = account.get_credentials_path()

    if user:
      print('SEND USER FOUND')

      # initialize project
      project.initialize(_user=user)

      # load template
      with open(kwargs['template'], 'r') as json_file:
        email = EmailTemplate(json.load(json_file))

      # loop through accounts
      for account in accounts:
        # if account is given only do that one
        if kwargs['email_to'] is None or account.email == kwargs['email_to']:
          if account.email in kwargs['ignore']:
            print('IGNORING: ', account.email)
          else:
            print('EMAILING: ', account.email)

            if kwargs['test']:
              # write to STDOUT
              print(email.get_html())
            else:
              # send message via email
              send_email('user', account.email, kwargs['email_from'], None,
                         email.get_subject(), email.get_text(),
                         email.get_html())
              sleep(1)
